# %%
# # Tutorial 5 venco.py

# This tutorial aims to give a more in depth overview into the DiaryBuilder
# class and showcases some features that can be customised. This tutorial will
# be added soon.



__maintainer__ = "Niklas Wulff, Fabia Miorelli"
__license__ = "BSD-3-Clause"

import math
import logging

from pathlib import Path
from typing import Union

import pandas as pd

from ..utils.metadata import read_metadata_config, write_out_metadata
from ..utils.utils import create_file_name, write_out, calculate_daily_seasonal_factor


class FlexEstimator:
    def __init__(self, configs: dict, data: pd.DataFrame):
        """
        In the Flexestimator, the previously defined activities are calculated
        one after the other. A further iteration loop is executed in the outer
        iteration cycle iterative_battrey_calculation. In the inner iteration
        loop, each activity id is processed one after the other. Firstly in a
        maximum consideration, which implies that as much as possible is
        always charged according to the usage profile. The minimum profile
        is then determined, which simulates a utilisation profile in which
        only as much is charged as is needed for the next planned trips.
        The two profiles and the resulting difference in battery level serve
        as a prerequisite for starting the next outer iteration loop and
        calling up the maximum and minimum profile with adjustment of the
        start variables. The __get_delta function is used to calculate both
        a max_delta and a min_delta depending on the start/end line of the
        min/max_battery_leve_start/end. As soon as the two delta values are
        below the selected epsilon value, the iteration is interrupted.
        Finally, the auxilliary_fuel_need is calculated in the flexestimator
        from the residual_need calculated for each trip and an output
        is generated.
        More detailed information on the different functions can be found in
        the documentation.

        Args:
            configs (dict): A dictionary containing a user_config_dictionary
            and a dev_config_dictionary
            activities (pd.DataFrame): a dataframe containing all trip and
            parking activities
        """
        self.dataset = configs["user_config"]["global"]["dataset"]
        self.user_config = configs["user_config"]
        self.dev_config = configs["dev_config"]
        self.electric_consumption = self.user_config[
            "flexestimators"]["electric_consumption"]["general"]
        self.season = self.user_config[
            "global"][
                "consider_temperature_cycle_dependency"]['season']
        self.seasons = ["winter", "spring", "summer", "fall"]
        self.temperature_factors = self.user_config[
            "global"][
                "consider_temperature_cycle_dependency"]["day_night_factors"]
        self.upper_battery_level = (
            self.user_config["flexestimators"]["battery_capacity"]
            * self.user_config["flexestimators"]["maximum_soc"]
        )
        self.lower_battery_level = (
            self.user_config["flexestimators"]["battery_capacity"]
            * self.user_config["flexestimators"]["minimum_soc"]
        )
        self.CCCV = self.user_config["flexestimators"]["CCCV_limit"]
        self.data = data
        self.activities = data['activities'].copy()
        self.is_trip = ~self.activities["trip_id"].isna()
        self.is_park = ~self.activities["park_id"].isna()
        self.is_first_activity = (
            self.activities["is_first_activity"].fillna(0).astype(bool)
        )
        self.is_last_activity = (
            self.activities["is_last_activity"].fillna(0).astype(bool)
        )
        self.activities[
            [
                "max_battery_level_start",
                "max_battery_level_end",
                "min_battery_level_start",
                "min_battery_level_end",
                "max_battery_level_end_unlimited",
                "uncontrolled_charging",
                "timestamp_end_uncontrolled_charging_unlimited",
                "timestamp_end_uncontrolled_charging",
                "min_battery_level_end_unlimited",
                "max_residual_need",
                "min_residual_need",
                "max_overshoot",
                "min_undershoot",
                "max_auxiliary_fuel_need",
                "min_auxiliary_fuel_need",
                "electric_consumption"
            ]
        ] = None
        self.activities_without_residual = None

    def _drain(self):
        """
        This function calculates the electric consumption first based on the
        seasonal factor and after that the drain of a specific trip according
        to its length based on the season (if enabled).
        """
        if self.user_config[
                "global"][
                "consider_temperature_cycle_dependency"][
                "annual"] & (self.season != 'None'):
            if self.user_config[
                    "global"][
                        "consider_temperature_cycle_dependency"]["daily"]:
                self.activities[
                    'electric_consumption'] = self.activities.apply(
                    lambda row:
                    self.electric_consumption *
                    (1 + self.user_config[
                        "flexestimators"]["electric_consumption"].get(
                        row['season'] + "_factor") *
                        calculate_daily_seasonal_factor(
                            self.user_config,
                            row['season'],
                            self.user_config[
                                "flexestimators"]["electric_consumption"],
                            row['trip_start_hour']
                            )),
                    axis=1
                )
            else:
                self.activities['electric_consumption'] = self.activities[
                    'season'].apply(
                    lambda season:
                    self.electric_consumption * (
                        1 + self.user_config[
                            "flexestimators"]["electric_consumption"].get(
                            season + "_factor"
                            ))
                       )
        elif (self.user_config["global"]["consider_temperature_cycle_dependency"]["season"] in self.seasons):
            if self.user_config[
                    "global"][
                        "consider_temperature_cycle_dependency"]["daily"]:
                self.activities[
                    'electric_consumption'] = self.activities.apply(
                    lambda row:
                    self.electric_consumption *
                    (1 + self.user_config[
                        "flexestimators"]["electric_consumption"].get(
                        row['season'] + "_factor") *
                        calculate_daily_seasonal_factor(
                            self.user_config,
                            self.season,
                            self.user_config[
                                "flexestimators"]["electric_consumption"],
                            row['trip_start_hour']
                            )),
                    axis=1
                )
            else:
                self.activities[
                    'electric_consumption'] = self.electric_consumption * (
                            1 + self.user_config[
                                "flexestimators"]["electric_consumption"].get(
                                self.season + "_factor"
                                ))
        else:
            self.activities['electric_consumption'] = self.electric_consumption

        self.activities["drain"] = (
            self.activities['trip_distance']
            * self.activities['electric_consumption'] / 100
            )

    def _max_charge_volume_first_parking_activity(self, first_park_activities):
        """
        Filters for the first parking activity and assigns a maximum charge
        volume based on the start soc

        Args:
            first_park_activities (dataframe): all rows of the dataframe that
            are first parking activities

        Returns:
            dataframe: dataframe wtith first activities
        """
        first_park_activities[
            "max_charge_volume"] = first_park_activities.apply(
                lambda row: self._max_charge_volume_per_parking_activity(
                            self.CCCV,
                            self.user_config["flexestimators"]["start_soc"],
                            self.user_config[
                                "flexestimators"]["battery_capacity"],
                            row["available_power"],
                            row["time_delta"],
                            self.user_config["flexestimators"]["time_steps"]),
                axis=1
                )

        return first_park_activities

    def _min_charge_volume_last_parking_activity(self, last_park_activities):
        """
        Filters for the last activities and calculates the min_charge_volume,
        based on the minimum soc,
        that would be the soc when arriving home with the min charge
        profile behaviour

        Args:
            last_park_activities (dataframe):  all rows of the dataframe that
            are last parking activities

        Returns:
            dataframe: dataframe wtith last activities
        """
        # NOTE: macht eigentlich keinen Sinn, weil erster charge vorgang immer
        # Laden ist und für den ersten Trip nicht zwei mal geladen werden muss
        last_park_activities["min_charge_volume"] = last_park_activities.apply(
                lambda row: self._min_charge_volume_per_parking_activity(
                            self.CCCV,
                            self.user_config["flexestimators"]["minimum_soc"],
                            self.user_config["flexestimators"]["minimum_soc"],
                            self.user_config[
                                "flexestimators"]["battery_capacity"],
                            row["available_power"],
                            row["time_delta"],
                            self.user_config["flexestimators"]["time_steps"]),
                axis=1
                )

        return last_park_activities

    @staticmethod
    def _max_charge_volume_per_parking_activity(CCCV,
                                                soc,
                                                battery_size,
                                                available_power,
                                                time,
                                                time_steps):
        """
        Calculates the max charge volume, related to the start_soc.
        If the soc_start is above 0.75, charging takes place slower

        Args:
            soc (tuple): start_soc (based on previous_trip_soc_end)
            available_power (_type_): available power from grid
            time (_type_): time delta for charging

        Returns:
            max_cahrge_volume: biggest possible charging_volume based on the
            charging time and available power
        """

        time_step = time / pd.Timedelta("1 hour")/time_steps

        if available_power == 0.0:
            max_charge_volume = 0
            return max_charge_volume

        elif soc < CCCV:
            # Calculate t, when soc = soc,cccv
            soc_CC = CCCV - soc
            time_switch = soc_CC * (
                battery_size / available_power
                ) * pd.Timedelta("1 hour")
            if time_switch < time:
                max_charge_volume = (
                    available_power
                    * time_switch
                    / pd.Timedelta("1 hour")
                ) + FlexEstimator._max_charge_volume_per_parking_activity(
                    CCCV, CCCV,
                    battery_size, available_power,
                    time-time_switch, time_steps)
            else:
                max_charge_volume = (
                    available_power
                    * time
                    / pd.Timedelta("1 hour")
                )

        elif soc >= CCCV:

            # to give a start max_charge_volume:
            max_charge_volume = 0
            for i in range(0, time_steps):

                soc = soc + (
                    available_power *
                    (-4 * soc + 4) * time_step
                ) / battery_size

                # to catch very long parking ttrips that are overestimated
                # with soc above 100% after first iteration
                # max_charging_volume will be the maximum possible charge
                # volume if soc = 1 so iteration can end
                if soc > 1:
                    soc = 1
                    max_charge_volume = battery_size * (1-CCCV)
                    break
                max_charge_volume = max_charge_volume + (
                    available_power *
                    (-4 * soc + 4) * time_step
                )
            # NOTE: charging above 100% no longer possible
            # -> no unlimited charge_volume
            
        return max_charge_volume

    @staticmethod
    def _min_charge_volume_per_parking_activity(CCCV,
                                                soc_min,
                                                soc_end,
                                                battery_size,
                                                available_power,
                                                time,
                                                time_steps):
        """
        Calculates a "minimum" charge volume based on the min_soc_start from
        the the next trip. min_soc_start_trip becomes soc_end of the parking
        activity. Because soc_start of a parking activity is only
        related to known values, it can be calculated to assign a slower
        or faster charge behaviour.

        Args:
            soc_min (_type_): minimum_soc (based on user_config)
            soc_end (_type_): soc on the end of previous trip activity
            available_power (_type_): available power from grid
            time (_type_): time delta for charging

        Returns:
            min_cahrge_volume: smallest possible charging_volume based on the
            next trip needed charging
        """
        if available_power == 0.0 or soc_end == soc_min:
            min_charge_volume = 0
            return min_charge_volume

        elif round(
                (soc_end - ((time / pd.Timedelta("1 hour")
                             ) * available_power / battery_size)), 2) < CCCV:
            soc_start = soc_end - (
                time / pd.Timedelta("1 hour") * available_power / battery_size)

            if soc_start < soc_min:
                soc_start = soc_min

            # switching from CC to CV happening within the parking time
            # calculate switching point
            soc_CC = CCCV - soc_start
            time_switch = soc_CC * (
                battery_size / available_power
                ) * pd.Timedelta("1 hour")

            if time_switch < time and soc_end > CCCV:
                CV_time = time - time_switch
                # switching happens within the given time
                min_charge_volume = (
                        available_power
                        * time_switch
                        / pd.Timedelta("1 hour")
                    ) + FlexEstimator._min_charge_volume_per_parking_activity(
                        CCCV, soc_min, soc_end,
                        battery_size, available_power,
                        CV_time, time_steps)
            else:
                min_charge_volume = (
                        available_power
                        * time
                        / pd.Timedelta("1 hour")
                    )

        else:
            time_step = time / pd.Timedelta("1 hour")/time_steps

            for i in range(0, time_steps):
                soc_start = (
                    soc_end - 4 * time_step * available_power / battery_size
                    ) / (-4 * time_step * available_power / battery_size + 1)
                soc_end = soc_start

            # to give a start min_charge_volume:
            min_charge_volume = 0
            for i in range(0, time_steps):
                soc_start = soc_start + (
                    available_power *
                    (-4 * soc_start + 4) * time_step
                    ) / battery_size

                min_charge_volume = min_charge_volume + (
                    available_power *
                    (-4 * soc_start + 4) * time_step
                    )

        return min_charge_volume

    def __battery_level_max(self, start_level: float) -> pd.Series:
        """
        Calculate the maximum battery level at the beginning and end of each
        activity. This represents the case of vehicle users always connecting
        when charging is available and charging as soon as possible as fast as
        possible until the maximum battery capacity is reached. act_temp is the
        overall collector for each activity's park and trip results, that will
        then get written to self.activities at the very end.

        Args:
            start_level (float): Battery start level for first activity of the
            activity chain

        Returns:
            pd.Series: activities dataframe with updated values
            for battery levels
        """
        logging.info("Starting maximum battery level calculation.")
        logging.info("Calculating maximum battery level for first activities.")

        act_temp = self.__first_activities(start_level=start_level)

        # Start and end for all trips and parkings in between
        set_acts = range(1, int(self.activities["park_id"].max()) + 1)
        subset_trip_activities = pd.DataFrame()  # Redundant?
        for act in set_acts:  # implementable via groupby with
            # actIDs as groups?
            logging.info(f"Calculating maximum battery level for act {act}.")
            trip_rows = (self.activities["trip_id"] == act) & (
                ~self.activities["is_first_activity"]
            )
            park_rows = (self.activities["park_id"] == act) & (
                ~self.activities["is_first_activity"]
            )
            trip_activities = self.activities.loc[trip_rows, :]
            park_activities = self.activities.loc[park_rows, :]
            # Filtering for the previous park activites that have the current
            # activity as next activity
            previous_park_activities = act_temp.loc[
                (act_temp["next_activity_id"] == act) & (
                    ~act_temp["park_id"].isna()), :
            ]

            subset_trip_activities = self.__calculate_max_battery_level_trip(
                activity_id=act,
                trip_activities=trip_activities,
                previous_park_activities=previous_park_activities,
            )

            act_temp = pd.concat([act_temp, subset_trip_activities],
                                 ignore_index=True)
            previous_trip_activities = act_temp.loc[
                (act_temp["next_activity_id"] == act) & (
                    ~act_temp["trip_id"].isna()), :
            ]
            subset_park_activities = self.__calculate_max_battery_level_park(
                activity_id=act,
                park_activities=park_activities,
                previous_trip_activities=previous_trip_activities,
            )

            act_temp = pd.concat([act_temp, subset_park_activities],
                                 ignore_index=True)
            # previous_trip_activities = subset_trip_activities  # Redundant?
        self.activities = act_temp.sort_values(
            by=["unique_id", "activity_id", "trip_id"]
        )
        return self.activities.loc[
            self.activities["is_last_activity"], [
                "unique_id", "max_battery_level_end"]
        ].set_index("unique_id")

    def __battery_level_min(self, end_level: pd.Series) -> pd.Series:
        """
        Calculate the minimum battery level at the beginning and end of each
        activity. This represents the case of vehicles just being charged for
        the energy required for the next trip and as late as possible. The loop
        works exactly inverted to the batteryLevelMax() function since later
        trips influence the energy that has to be charged in parking activities
        before. Thus, activities are looped over from the last activity to
        first.

        Args:
            end_level (pd.Series): A series containing the battery_end levels
            to start the battery_level_min

        Returns:
            pd.Series: The activities dataframe with updated
            batter_level_min_end/start values
        """
        logging.info("Starting minimum battery level calculation.")
        logging.info("Calculate minimum battery level for last activities.")
        last_activities = self.__last_activities(end_level=end_level)
        act_temp = last_activities
        # Start and end for all trips and parkings starting from the last
        # activities, then looping to earlier activities
        n_act = int(self.activities["park_id"].max())
        for act in range(n_act, -1, -1):
            logging.info(f"Calculate minimum battery level for act {act}.")
            trip_rows = (self.activities["trip_id"] == act) & (
                ~self.activities["is_last_activity"]
            )
            park_rows = (self.activities["park_id"] == act) & (
                ~self.activities["is_last_activity"]
            )
            trip_activities = self.activities.loc[trip_rows, :]
            park_activities = self.activities.loc[park_rows, :]

            next_trip_activities = act_temp.loc[
                (act_temp["previous_activity_id"] == act)
                & (~act_temp["trip_id"].isna()),
                :,
            ].copy()
            if act != n_act:
                next_trip_activities[
                    "park_activity_index"] = park_activities.index
                next_trip_activities = next_trip_activities.set_index(
                    "park_activity_index")
                next_trip_activities["soc_min_start"] = next_trip_activities[
                    "min_battery_level_start"] / self.user_config[
                        "flexestimators"]["battery_capacity"]

                if park_activities.empty == False:  # noqa: E712
                    subset_park_activities = self.__calculate_min_battery_level_park(
                        activity_id=act,
                        park_activities=park_activities,
                        next_trip_activities=next_trip_activities
                    )
                    act_temp = pd.concat(
                        [act_temp, subset_park_activities], ignore_index=True)

            next_park_activities = act_temp.loc[
                (act_temp["previous_activity_id"] == act)
                & (~act_temp["park_id"].isna()),
                :,
            ]
            subset_trip_activities = self.__calculate_min_battery_level_trip(
                activity_id=act,
                trip_activities=trip_activities,
                next_park_activities=next_park_activities,
            )
            act_temp = pd.concat(
                [act_temp, subset_trip_activities], ignore_index=True)
        self.activities = act_temp.sort_values(
            by=["unique_id", "activity_id", "trip_id"], ignore_index=True
        )
        return self.activities.loc[
            self.activities["is_first_activity"],
            ["unique_id", "min_battery_level_start"],
        ].set_index("unique_id")

    def __first_activities(
            self,
            start_level: Union[float, pd.Series]
            ) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        This function filters for all first activities and calls the
        _calculate_max_battery_level_first_activity function, that handels all
        first activities, parking as well as trips. If over_night_spliiting is
        activated there is an additional calculation of the battery level for
        parking activities that do not have the activity_id= 0. But due to
        overnight splitting they are first activities.

        Args:
            start_level (Union[float, pd.Series]): A float or Series that
            contains battery_levels used to start the activity calculation

        Returns:
            pd.DataFrame: updated dataframe that contains all handeld first
            activities
        """
        first_activities = self._calculate_max_battery_level_first_activity(
            start_level=start_level)
        first_parking_activities = first_activities.loc[
            ~first_activities["park_id"].isna(), :]
        first_trip_activities = first_activities.loc[(
            ~first_activities["trip_id"].isna()), :]
        first_park_activities_overnight = self.activities.loc[
                (self.activities['is_first_park_activity']) & ~(
                    self.activities['is_first_activity'])].copy()

        if self.user_config[
            'dataparsers'][
                'split_overnight_trips'
                ] & first_park_activities_overnight.empty is False:
            # Treat parking activities after morning split trips.
            # They have activitiy_id==0 but are the second activity.
            first_park_activities_overnight = self.__calculate_max_battery_level_park(
                activity_id=0,
                park_activities=first_park_activities_overnight,
                previous_trip_activities=first_trip_activities,
            )
            return pd.concat(
                [
                    first_parking_activities,
                    first_trip_activities,
                    first_park_activities_overnight,
                ]
            )
        else:
            return pd.concat([first_parking_activities, first_trip_activities])

    def _calculate_max_battery_level_first_activity(
        self, start_level: Union[float, pd.Series]
    ) -> pd.DataFrame:
        """
        Calculate maximum battery levels at beginning and end of the first
        activities. This includes first park
        activities (inferred) before first MiD trips as well as morning splits
        first trip activities. Parking activities
        after morning split trips are treated later in [ELSEWHERE].

        Args:
            start_level (Union[float, pd.Series]): Start battery level at
            beginning of simulation (MON, 00:00). Defaults
            to self.upper_battery_level, the maximum battery level.
            In iterations, this is a pandas Series with the
            respective unique_ids in the index.

        Returns:
            pd.DataFrame: First activities with all battery level columns as
            anchor for the consecutive calculation of maximum charge
        """
        # First activities - parking and trips
        first_activities = self.activities.loc[
            self.activities["is_first_activity"], :
        ].copy()

        # Needed for setting start_level at correct unique_id in iterations
        first_activities = first_activities.set_index("unique_id")
        first_activities["max_battery_level_start"] = start_level
        first_activities = first_activities.reset_index()
        first_trip_activities = self.__calc_first_trip_activities(
            first_trip_activities=first_activities.loc[
                ~first_activities["trip_id"].isna(), :].copy())

        first_activities = self._max_charge_volume_first_parking_activity(
            first_park_activities=first_activities.loc[
                ~first_activities["park_id"].isna()].copy())

        first_park_activities = self.__calc_first_park_activities(
            first_park_activities=first_activities.loc[
                ~first_activities["park_id"].isna()].copy())

        first_activities = pd.concat(
            [first_park_activities, first_trip_activities])
        return first_activities.sort_values(
            by=["unique_id"]
        )  # Only one activity per unique_id exists

    def __calc_first_park_activities(
        self, first_park_activities: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Function to handle all first parking activities. Calculates
        max_battery_level_end of this parking activity by adding the
        max_charge_volume, and the max_battery_level_end_unlimited by
        ignoring the maximum battery capycity.

        Args:
            first_park_activities (pd.DataFrame): A dataframe containing only
            first park activities from the activities dataframe

        Returns:
            pd.DataFrame: dataframe containing only first park activities with
            new updated max_battery_levels
        """
        first_park_activities.loc[:, "max_battery_level_end_unlimited"] = (
            first_park_activities["max_battery_level_start"]
            + first_park_activities["max_charge_volume"]
        )
        first_park_activities.loc[
            :, "max_battery_level_end"
        ] = first_park_activities.loc[
            :, "max_battery_level_end_unlimited"].where(
            first_park_activities.loc[:, "max_battery_level_end_unlimited"]
            <= self.upper_battery_level,
            other=self.upper_battery_level,
        )
        first_park_activities.loc[:, "max_overshoot"] = (
            first_park_activities["max_battery_level_end_unlimited"]
            - first_park_activities["max_battery_level_end"]
        )

        return first_park_activities

    def __calc_first_trip_activities(
        self, first_trip_activities: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Function that handles all first activities being a trip
        (only happening with over night splitting). Function substracts the
        drain from every max_battery_level_start of the trip and calculates
        the max_battery_level_end as well as max_battery_level_end_unlimited
        if the drain exceeds the battery capacity and there fore residual
        need will result.

        Args:
            first_trip_activities (pd.DataFrame): dataframe containing only
            first trip activities

        Returns:
            pd.DataFrame: dataframe containing only first trip activities with
            new calculated max_battery_level_end
        """
        first_trip_activities.loc[:, "max_battery_level_end_unlimited"] = (
            first_trip_activities.loc[:, "max_battery_level_start"]
            - first_trip_activities.loc[:, "drain"]
        )
        first_trip_activities.loc[
            :, "max_battery_level_end"
        ] = first_trip_activities.loc[
            :, "max_battery_level_end_unlimited"].where(
            first_trip_activities.loc[:, "max_battery_level_end_unlimited"]
            >= self.lower_battery_level,
            other=self.lower_battery_level,
        )
        res = (
            first_trip_activities.loc[:, "max_battery_level_end"]
            - first_trip_activities.loc[:, "max_battery_level_end_unlimited"]
        )
        first_trip_activities.loc[:, "max_residual_need"] = res.where(
            first_trip_activities.loc[:, "max_battery_level_end_unlimited"]
            < self.lower_battery_level,
            other=0,
        )

        return first_trip_activities

    def __last_activities(self, end_level: pd.Series) -> pd.DataFrame:
        """
        Calculate the minimum battery levels for the last activity in the
        dataset determined by the maximum activity
        ID.

        Args:
            end_level (float or pd.Series): End battery level at end of
            simulation time (last_bin). Defaults to
            self.lower_battery_level, the minimum battery level. Can be either
            of type float (in first iteration) or
            pd.Series with respective unique_id in the index.

        Returns:
            pd.DataFrame: Activity data set with the battery variables set for
            all last activities of the activity chains
        """
        # Last activities - parking and trips
        last_activities_in = self.activities.loc[
            self.activities["is_last_activity"], :
        ].copy()
        is_trip = ~last_activities_in["trip_id"].isna()

        indeces_last_activities = last_activities_in.set_index("unique_id")
        indeces_last_activities["min_battery_level_end"] = end_level
        indeces_last_activities.loc[
            indeces_last_activities["trip_id"].isna(),
            "min_battery_level_start"
        ] = end_level  # For park activities

        self._min_charge_volume_last_parking_activity(
            indeces_last_activities.loc[
                indeces_last_activities["trip_id"].isna()])

        last_activities = indeces_last_activities.reset_index("unique_id")
        last_activities.index = last_activities_in.index

        last_activities.loc[is_trip, "min_battery_level_start_unlimited"] = (
            last_activities.loc[is_trip, "min_battery_level_end"]
            + last_activities.loc[is_trip, "drain"]
        )
        last_activities.loc[
            is_trip, "min_battery_level_start"] = last_activities.loc[
            is_trip, "min_battery_level_start_unlimited"
        ].where(
            last_activities.loc[is_trip, "min_battery_level_start_unlimited"]
            <= self.upper_battery_level,
            other=self.upper_battery_level,
        )
        residual_need = (
            last_activities.loc[is_trip, "min_battery_level_start_unlimited"]
            - self.upper_battery_level
        )
        last_activities.loc[
            is_trip, "min_residual_need"] = residual_need.where(
            residual_need >= 0, other=0
        ).astype(float)
        return last_activities

    def __calculate_max_battery_level_trip(
        self,
        activity_id: int,
        trip_activities: pd.DataFrame,
        previous_park_activities: pd.DataFrame = None,
    ) -> pd.DataFrame:
        """
        Function that handles activities with current trip_id= act. Function
        substracts the drain from every max_battery_level_start that is based
        on the previous parking (charging) activity of the trip and calculates
        the max_battery_level_end as well as max_battery_level_end_unlimited
        if the drain exceeds the battery capacity and there fore residual need
        will result.

        Args:
            activity_id (int): the current id of activity that is handeld
            trip_activities (pd.DataFrame): dataframe that contains all trips
            with trip_id = current act
            previous_park_activities (pd.DataFrame, optional):
            dataframe that stores all previous park activities that have
            same act_id as the current trip_id

        Returns:
            pd.DataFrame: _description_
        """
        # Index setting of trip activities to be updated
        active_unique_ids = trip_activities.loc[:, "unique_id"]
        multi_index_trip = [
            (id, activity_id, None) for id in active_unique_ids]
        indeces_trip_activities = trip_activities.set_index(
            ["unique_id", "trip_id", "park_id"]
        )
        # Index setting of previous park activities as basis for the update
        previous_park_ids = trip_activities.loc[:, "previous_activity_id"]
        multi_index_park = [
            (id, None, act) for id, act in zip(
                active_unique_ids, previous_park_ids)
        ]
        indeces_previous_park_activities = previous_park_activities.set_index(
            ["unique_id", "trip_id", "park_id"]
        )
        # Calculation of battery level at start and end of trip
        indeces_trip_activities.loc[
            multi_index_trip, "max_battery_level_start"] = (
            indeces_previous_park_activities.loc[
                multi_index_park, "max_battery_level_end"
            ].values
        )
        indeces_trip_activities.loc[
            multi_index_trip, "max_battery_level_end_unlimited"
        ] = (
            indeces_trip_activities.loc[
                multi_index_trip, "max_battery_level_start"]
            - indeces_trip_activities.loc[multi_index_trip, "drain"]
        )
        indeces_trip_activities.loc[
            multi_index_trip, "max_battery_level_end"
        ] = indeces_trip_activities.loc[
            multi_index_trip, "max_battery_level_end_unlimited"
        ].where(
            indeces_trip_activities.loc[
                multi_index_trip, "max_battery_level_end_unlimited"
            ]
            >= self.lower_battery_level,
            other=self.lower_battery_level,
        )
        res = (
            indeces_trip_activities.loc[
                multi_index_trip, "max_battery_level_end"]
            - indeces_trip_activities.loc[
                multi_index_trip, "max_battery_level_end_unlimited"
            ]
        )
        indeces_trip_activities.loc[
            multi_index_trip, "max_residual_need"] = res.where(
            indeces_trip_activities.loc[
                multi_index_trip, "max_battery_level_end_unlimited"
            ]
            < self.lower_battery_level,
            other=0,
        )
        return indeces_trip_activities.reset_index()

    def __calculate_min_battery_level_trip(
        self,
        activity_id: int,
        trip_activities: pd.DataFrame,
        next_park_activities: pd.DataFrame = None,
    ) -> pd.DataFrame:
        """
        Function that handles activities with current trip_id= act. Function
        adds the drain from on every min_battery_level_end, based on the
        min_battery_level_start of the next parking activity of the trip
        and calculates the min_battery_level_start as well as
        min_battery_level_start_unlimited if the drain exceeds the battery
        capacity and therefore residual need will result.

        Args:
            activity_id (int): _description_
            trip_activities (pd.DataFrame): _description_
            next_park_activities (pd.DataFrame, optional): _description_.
            Defaults to None.

        Returns:
            pd.DataFrame: _description_
        """
        active_unique_ids = trip_activities.loc[:, "unique_id"]
        multi_index_trip = [
            (id, activity_id, None) for id in active_unique_ids]
        indeces_trip_activities = trip_activities.set_index(
            ["unique_id", "trip_id", "park_id"]
        )

        next_park_ids = trip_activities.loc[:, "next_activity_id"]
        multi_index_park = [
            (id, None, act) for id, act in zip(
                active_unique_ids, next_park_ids)
        ]
        indeces_next_park_activities = next_park_activities.set_index(
            ["unique_id", "trip_id", "park_id"]
        )
        indeces_trip_activities.loc[
            multi_index_trip, "min_battery_level_end"] = (
            indeces_next_park_activities.loc[
                multi_index_park, "min_battery_level_start"
            ].values
        )

        # Setting minimum battery end level for trip
        indeces_trip_activities.loc[
            multi_index_trip, "min_battery_level_start_unlimited"
        ] = (
            indeces_trip_activities.loc[
                multi_index_trip, "min_battery_level_end"]
            + indeces_trip_activities.loc[multi_index_trip, "drain"]
        )
        indeces_trip_activities.loc[
            multi_index_trip, "min_battery_level_start"
        ] = indeces_trip_activities.loc[
            multi_index_trip, "min_battery_level_start_unlimited"
        ].where(
            indeces_trip_activities.loc[
                multi_index_trip, "min_battery_level_start_unlimited"
            ]
            <= self.upper_battery_level,
            other=self.upper_battery_level,
        )
        residual_need = (
            indeces_trip_activities.loc[
                multi_index_trip, "min_battery_level_start_unlimited"
            ]
            - self.upper_battery_level
        )
        indeces_trip_activities.loc[multi_index_trip, "min_residual_need"] = (
            residual_need.where(residual_need >= 0, other=0)
        )
        return indeces_trip_activities.reset_index()

    def __calculate_max_battery_level_park(
        self,
        activity_id: int,
        park_activities: pd.DataFrame,
        previous_trip_activities: pd.DataFrame = None,
    ) -> pd.DataFrame:
        """
        Calculate the maximum battery level of the given
        parking activities for the activity ID given by activity_id.
        Previous trip activities are used as boundary for
        max_battery_level_start. This function is called multiple
        times once per activity ID. It is then applied to all activities
        with the given activity ID in a vectorized
        manner.

        Args:
            activity_id (int): Activity ID in current loop
            park_activities (pd.DataFrame): _description_
            previous_trip_activities (pd.DataFrame, optional): _description_.
            Defaults to None.

        Returns:
            pd.DataFrame: Park activities with maximum battery level columns.
        """

        park_activities = park_activities.set_index(["unique_id"])
        indeces_previous_trip_activities = previous_trip_activities.set_index(
            ["unique_id"])
        park_activities["soc_max_start"] = indeces_previous_trip_activities[
            "max_battery_level_end"] / self.upper_battery_level
        park_activities["max_charge_volume"] = park_activities.apply(
                lambda row: self._max_charge_volume_per_parking_activity(
                        self.CCCV,
                        row["soc_max_start"],
                        self.upper_battery_level,
                        row["available_power"],
                        row["time_delta"],
                        self.user_config["flexestimators"]["time_steps"]
                        ), axis=1
                )

        park_activities = park_activities.reset_index()
        # Index setting of park activities to be updated
        active_unique_ids = park_activities.loc[:, "unique_id"]
        multi_index_park = [
            (id, None, activity_id) for id in active_unique_ids]
        indeces_park_activities = park_activities.set_index(
            ["unique_id", "trip_id", "park_id"]
        )

        # Index setting of previous trip activities used to update
        previous_trip_ids = park_activities.loc[:, "previous_activity_id"]
        multi_index_trip = [
            (id, act, None) for id, act in zip(active_unique_ids,
                                               previous_trip_ids)
        ]
        indeces_previous_trip_activities = previous_trip_activities.set_index(
            ["unique_id", "trip_id", "park_id"]
        )

        # Calculation of battery level at start and end of park activity
        indeces_park_activities.loc[
            multi_index_park, "max_battery_level_start"] = (
            indeces_previous_trip_activities.loc[
                multi_index_trip, "max_battery_level_end"
            ].values
        )
        indeces_park_activities["max_battery_level_end_unlimited"] = (
            indeces_park_activities.loc[
                multi_index_park, "max_battery_level_start"]
            + indeces_park_activities.loc[
                multi_index_park, "max_charge_volume"]
        )
        indeces_park_activities.loc[
            multi_index_park, "max_battery_level_end"
        ] = indeces_park_activities["max_battery_level_end_unlimited"].where(
            indeces_park_activities["max_battery_level_end_unlimited"]
            <= self.upper_battery_level,
            other=self.upper_battery_level,
        )
        temporary_overshoot = (
            indeces_park_activities["max_battery_level_end_unlimited"]
            - self.upper_battery_level
        )
        indeces_park_activities["max_overshoot"] = temporary_overshoot.where(
            temporary_overshoot >= 0, other=0
        )
        return indeces_park_activities.reset_index()

    def __calculate_min_battery_level_park(
        self,
        activity_id: int,
        park_activities: pd.DataFrame,
        next_trip_activities: pd.DataFrame = None,
    ) -> pd.DataFrame:
        """
        Calculate minimum battery levels for given parking activities based on
        the given next trip activities.
        The calculated battery levels only suffice for the trips and thus
        describe a technical lower level for
        each activity. This function is called looping through the parking
        activities from largest to smallest.
        The column "minOvershoot" describes electricity volume that can be
        charged beyond the given batterycapacity.

        Args:
            activity_id (int): _description_
            park_activities (pd.DataFrame): _description_
            next_trip_activities (pd.DataFrame, optional): _description_.
            Defaults to None.

        Returns:
            _type_: _description_
        """
        # match next_trip_index unique_id with park_activities index unique_id
        park_activities = park_activities.set_index(["unique_id"])
        next_trip_activities = next_trip_activities.set_index(["unique_id"])

        park_activities["soc_min_end"] = next_trip_activities[
            "min_battery_level_start"]/self.upper_battery_level
        park_activities["min_charge_volume"] = park_activities.apply(
                lambda row: self._min_charge_volume_per_parking_activity(
                        self.CCCV,
                        self.user_config["flexestimators"]["minimum_soc"],
                        row["soc_min_end"],
                        self.upper_battery_level,
                        row["available_power"],
                        row["time_delta"],
                        self.user_config["flexestimators"]["time_steps"]
                        ), axis=1
            )

        park_activities = park_activities.reset_index()
        next_trip_activities = next_trip_activities.reset_index()
        # Composing park activity index to be set
        active_unique_ids = park_activities.loc[:, "unique_id"]
        multi_index_park = [
            (id, None, activity_id) for id in active_unique_ids]
        indeces_park_activities = park_activities.set_index(
            ["unique_id", "trip_id", "park_id"]
        )
        # Composing trip activity index to get battery level from
        next_trip_ids = park_activities.loc[:, "next_activity_id"]
        multi_index_trip = [
            (id, act, None) for id, act in zip(
                active_unique_ids,
                next_trip_ids
                )
        ]
        indeces_next_trip_activities = next_trip_activities.set_index(
            ["unique_id", "trip_id", "park_id"]
        )

        indeces_park_activities.loc[
            multi_index_park,
            "min_battery_level_end"] = (
            indeces_next_trip_activities.loc[
                multi_index_trip, "min_battery_level_start"
            ].values
        )
        indeces_park_activities["min_battery_level_start_unlimited"] = (
                indeces_park_activities.loc[
                    multi_index_park,
                    "min_battery_level_end"]
                - indeces_park_activities.loc[
                    multi_index_park,
                    "min_charge_volume"]
        )
        indeces_park_activities.loc[
            multi_index_park, "min_battery_level_start"
        ] = indeces_park_activities["min_battery_level_start_unlimited"].where(
            indeces_park_activities["min_battery_level_start_unlimited"]
            >= self.lower_battery_level,
            other=self.lower_battery_level,
        )
        temporary_undershoot = (
            indeces_park_activities["min_battery_level_start_unlimited"]
            - self.lower_battery_level
        )
        indeces_park_activities["min_undershoot"] = temporary_undershoot.where(
            temporary_undershoot >= 0, other=0
        )
        return indeces_park_activities.reset_index()

    def _uncontrolled_charging(self):
        """
        calculates the difference between the start and end level of a battery
        resuting in the realistic value of energy
        that was charged into the battery
        """
        park_activities = self.activities.loc[
            self.activities["trip_id"].isna(), :
        ].copy()
        park_activities["uncontrolled_charging"] = (
            park_activities["max_battery_level_end"]
            - park_activities["max_battery_level_start"]
        )
        # Calculate timestamp at which charging ends disregarding parking end
        park_activities[
            "timestamp_end_uncontrolled_charging_unlimited"
            ] = park_activities.apply(
            lambda x: self._calculate_charging_end_timestamp(
                start_timestamp=x["timestamp_start"],
                start_battery_level=x["max_battery_level_start"],
                power=x["available_power"],
            ),
            axis=1,
        )

        # Take into account possible earlier disconnection due
        # to end of parking
        park_activities[
            "timestamp_end_uncontrolled_charging"
            ] = park_activities[
            "timestamp_end_uncontrolled_charging_unlimited"
        ].where(
            park_activities["timestamp_end_uncontrolled_charging_unlimited"]
            <= park_activities["timestamp_end"],
            other=park_activities["timestamp_end"],
        )

        self.activities.loc[
            self.activities["trip_id"].isna(), :] = park_activities

    def _calculate_charging_end_timestamp(
        self,
        start_timestamp: pd.Timestamp,
        start_battery_level: float,
        power: float
    ) -> pd.Timestamp:
        """
        Function calculates an end_timp_stamp where the realistic charging has
        stopped for the specific parking act. This can be due to the battery
        being full or the vehcile taking its next trip and is therefore being
        plugged off the grid.

        Args:
            start_timestamp (pd.Timestamp): timestamp where the charging
            process begins
            start_battery_level (float): battery level at te start timestamp
            power (float): available power for the individual parking activity
            based on the grid

        Returns:
            pd.Timestamp: dataframe containing all end timestamps for each
            charging activity
        """
        #TODO: check if correct
        if power == 0:
            return pd.NA
        else:
            delta_battery_level = self.upper_battery_level-start_battery_level

            if start_battery_level/self.user_config["flexestimators"][
                    "battery_capacity"] < self.CCCV:
                time_for_charge = delta_battery_level / power  # in hours
            else:
                time_for_charge = delta_battery_level / (
                    power * (
                        -4 * (
                            start_battery_level/self.user_config[
                                "flexestimators"]["battery_capacity"]
                                ) + 4
                                )
                            )
            return start_timestamp + pd.Timedelta(value=time_for_charge,
                                                  unit="h").round(freq="s")

    def _auxiliary_fuel_need(self):
        """
        Function calculates the auxiliary fuel needed for the trips that
        cannot be fulfilled purely by energy stored in
        the battery. E.g. if the sum of all trips requires 55 kWh and there
        is no charging available but the battery
        capacity is set to 50 kWh, 5 kWh will be required from auxiliary fuel.
        The difference between the variables
        max_auxiliary_fuel_need and min_auxiliary_fuel_need is negligible and
        occurs in edge cases with very specific
        chains of charging availabilities and trips. If the option
        "filter_fuel_need" is set to True in the
        flexestimator section of the user_config, both variables are used in
        FlexEstimator._filter_residual_need()
        to drop trips where auxiliary fuel is needed.
        """
        self.activities["max_auxiliary_fuel_need"] = (
            self.activities["max_residual_need"]
            * self.user_config["flexestimators"]["fuel_consumption"]
            / self.activities["electric_consumption"]
        )

        self.activities["min_auxiliary_fuel_need"] = (
            self.activities["min_residual_need"]
            * self.user_config["flexestimators"]["fuel_consumption"]
            / self.activities["electric_consumption"]
        )

    def _filter_residual_need(
        self, activities: pd.DataFrame, index_columns: list
    ) -> pd.DataFrame:
        """
        Filter out days (uniqueIDs) that require additional fuel, i.e. for
        which the trip distance cannot be
        completely be fulfilled with the available charging power.
        Since additional fuel for a single trip motivates
        filtering out the whole vehicle, index_columns defines the columns
        that make up one vehicle. If index_columns is
        ['unique_id'], all uniqueIDs that have at least one trip requiring
        fuel are disregarded. If index_columns is
        ['category_id', 'week_id'] each unique combination of category_id and
        week_id (each "week") for which fuel is
        required in at least one trip is disregarded.

        Args:
            activities (pd.DataFrame): Activities data set containing at least
            the columns 'unique_id' and
                'max_residual_need'
            index_columns (list): Columns that define a "day", i.e. all unique
            combinations where at least one activity
                requires residual fuel are disregarded.
        """
        indeces_activities = activities.set_index(index_columns)
        indeces_out = (~indeces_activities["max_residual_need"].isin([None, 0]
                                                                     )) | (
            ~indeces_activities["min_residual_need"].isin([None, 0])
        )

        if len(index_columns) == 1:
            category_week_ids_out = indeces_activities.index[indeces_out]
            activities_filter = indeces_activities.loc[
                ~indeces_activities.index.isin(category_week_ids_out)
            ]
        else:
            category_week_ids_out = activities.loc[indeces_out.values,
                                                   index_columns]
            filter = category_week_ids_out.apply(
                lambda x: tuple(x), axis=1).unique()
            activities_filter = indeces_activities.loc[
                ~indeces_activities.index.isin(filter), :
            ]
        return activities_filter.reset_index()

    def __write_output(self):
        """
        Writes the output of the flexestimator calculation to the specified
        disk path
        """
        if self.user_config["global"]["write_output_to_disk"]["flex_output"]:
            root = Path(self.user_config[
                "global"]["absolute_path"]["vencopy_root"])
            folder = self.dev_config["global"]["relative_path"]["flex_output"]
            file_name = create_file_name(
                user_config=self.user_config,
                dev_config=self.dev_config,
                file_name_id="output_flexestimator",
                dataset=self.dataset,
            )
            write_out(data=self.activities, path=root / folder / file_name)
            self._write_metadata(file_name=root / folder / file_name)

    def generate_metadata(self, metadata_config, file_name):
        metadata_config["name"] = file_name
        metadata_config["title"] = "National Travel Survey activities dataframe"
        metadata_config["description"] = "Trips and parking activities from venco.py including profiles representing the available charging power, an uncontrolled charging profile, the battery drain, and the maximum and minum battery level."
        metadata_config["sources"] = [f for f in metadata_config["sources"] if f["title"] in self.dataset]
        reference_resource = metadata_config["resources"][0]
        this_resource = reference_resource.copy()
        this_resource["name"] = file_name.rstrip(".csv")
        this_resource["path"] = file_name
        these_fields = [f for f in reference_resource["schema"][self.dataset]["fields"]["flexestimators"] if f["name"] in self.activities.columns]
        this_resource["schema"] = {"fields": these_fields}
        metadata_config["resources"].pop()
        metadata_config["resources"].append(this_resource)
        return metadata_config

    def _write_metadata(self, file_name):
        metadata_config = read_metadata_config()
        class_metadata = self.generate_metadata(metadata_config=metadata_config, file_name=file_name.name)
        write_out_metadata(metadata_yaml=class_metadata, file_name=file_name.as_posix().replace(".csv", ".metadata.yaml"))

    def __iterative_battery_level_calculation(
        self,
        max_iteration: int,
        epsilon: float,
        battery_capacity: float,
        number_vehicles: int,
    ):
        """
        A single iteration of calculation maximum battery levels, uncontrolled charging and minimum battery levels
        for each trip. Initial battery level for first iteration loop per unique_id in index. Start battery level will
        be set to end battery level consecutively. Function operates on class attribute self.activities.

        Args:
            max_iteration (int): Maximum iteration limit if epsilon threshold is never reached.
            epsilon (float): Share of total aggregated battery fleet capacity (e.g. 0.01 for 1% would relate to a
                threshold of 100 Wh per car for a 10 kWh battery capacity.)
            battery_capacity (float): Average nominal battery capacity per vehicle in kWh.
            number_vehicles (int): Number of vehicles in the empiric mobility pattern data set.
        """
        max_battery_level_end = (
            self.upper_battery_level * self.user_config["flexestimators"]["start_soc"]
        )
        min_battery_level_start = self.lower_battery_level
        absolute_epsilon = int(
            self.__absolute_epsilon(
                epsilon=epsilon,
                battery_capacity=battery_capacity,
                number_vehicles=number_vehicles,
            )
        )

        max_battery_level_end = self.__battery_level_max(
            start_level=max_battery_level_end
        )
        self._uncontrolled_charging()
        min_battery_level_start = self.__battery_level_min(
            end_level=min_battery_level_start
        )

        max_delta = self.__get_delta(
            start_column="max_battery_level_start", end_column="max_battery_level_end"
        )
        min_delta = self.__get_delta(
            start_column="min_battery_level_start", end_column="min_battery_level_end"
        )

        logging.info(
            f"Finished iteration {1} / {max_iteration}. Delta max battery level is {int(max_delta)}, delta min "
            f"battery level is {int(min_delta)} and threshold epsilon is {absolute_epsilon}."
        )

        for i in range(1, max_iteration + 1):
            if max_delta < absolute_epsilon and min_delta < absolute_epsilon:
                break

            elif max_delta >= absolute_epsilon:
                max_battery_level_end = self.__battery_level_max(
                    start_level=max_battery_level_end
                )
                self._uncontrolled_charging()
                max_delta = self.__get_delta(
                    start_column="max_battery_level_start",
                    end_column="max_battery_level_end",
                )

            else:
                min_battery_level_start = self.__battery_level_min(
                    end_level=min_battery_level_start
                )
                min_delta = self.__get_delta(
                    start_column="min_battery_level_start",
                    end_column="min_battery_level_end",
                )

            logging.info(
                f"Finished iteration {i} / {max_iteration}. Delta max battery level is {int(max_delta)}, delta min "
                f"battery level is {int(min_delta)} and threshold epsilon is {absolute_epsilon}."
            )

    def __absolute_epsilon(
        self, epsilon: float, battery_capacity: float, number_vehicles: int
    ) -> float:
        """
        Calculates the absolute threshold of battery level deviatiation (delta in kWh for the whole fleet)
        used for interrupting the battery level calculation iterations.

        Args:
            epsilon (float): Share of total aggregated battery fleet capacity (e.g. 0.01 for 1% would relate to a
                threshold of 100 Wh per car for a 10 kWh battery capacity.)
            batteryCapacity (float): Average battery capacity per car
            number_vehicles (int): Number of vehicles

        Returns:
            float: Absolute iteration threshold in kWh of fleet battery
        """
        return epsilon * battery_capacity * number_vehicles

    def __get_delta(self, start_column: str, end_column: str) -> float:
        """
        Function calculates an absolute delta from the column max/min_battery_level_end of all last activies and the column max/min_battery_level_start of all first activities.

        Args:
            start_column (str): name of start_column to call (max/min_battery_level_start) 
            end_column (str): name of end column to call (max/min_battery_level_end)

        Returns:
            float: absolute delta
        """
        delta = abs(
            self.activities.loc[self.activities["is_last_activity"], end_column].values
            - self.activities.loc[
                self.activities["is_first_activity"], start_column
            ].values
        ).sum()
        return delta

    def estimate_technical_flexibility_no_boundary_constraints(self) -> pd.DataFrame:
        """
        Main run function for the class WeekFlexEstimator. Calculates uncontrolled charging as well as technical
        boundary constraints for controlled charging and feeding electricity back into the grid on an indvidiual vehicle
        basis. If filter_fuel_need is True, only electrifiable days are considered.

        Returns:
            pd.DataFrame: Activities data set comprising uncontrolled charging and flexible charging constraints for
            each car.
        """
        self._drain()
        self.__battery_level_max(start_level=self.upper_battery_level * self.user_config["flexestimators"]["start_soc"])
        self._uncontrolled_charging()
        self.__battery_level_min()
        self._auxiliary_fuel_need()
        if self.user_config["flexestimators"]["filter_fuel_need"]:
            self.activities = self._filter_residual_need(
                activities=self.activities, index_columns=["unique_id"]
            )
        if self.user_config["global"]["write_output_to_disk"]["flex_output"]:
            self.__write_output()
        logging.info("Technical flexibility estimation ended.")
        return self.activities

    @staticmethod
    def _cleanup_dataset(activities):
        activities.drop(
            columns=['max_battery_level_end',
                     'min_battery_level_start',
                     'max_battery_level_end_unlimited',
                     'max_battery_level_end_unlimited',
                     'timestamp_end_uncontrolled_charging_unlimited',
                     'min_battery_level_end_unlimited',
                     'min_battery_level_end_unlimited',
                     'max_residual_need',
                     'min_residual_need',
                     'max_overshoot',
                     'min_undershoot',
                     # 'auxiliary_fuel_need',
                     'max_charge_volume',
                     'min_battery_level_start_unlimited'], inplace=True)
        return activities

    def estimate_technical_flexibility_through_iteration(self) -> pd.DataFrame:
        """
        Main run function for the class WeekFlexEstimator. Calculates uncontrolled charging as well as technical
        boundary constraints for controlled charging and feeding electricity back into the grid on an indvidiual vehicle
        basis. If filter_fuel_need is True, only electrifiable days are considered.

        Returns:
            pd.DataFrame: Activities data set comprising uncontrolled charging and flexible charging constraints for
            each car.
        """
        self._drain()
        self.__iterative_battery_level_calculation(
            max_iteration=self.user_config["flexestimators"]["max_iterations"],
            epsilon=self.user_config["flexestimators"]["epsilon_battery_level"],
            battery_capacity=self.upper_battery_level,
            number_vehicles=len(self.activities["unique_id"].unique()),
        )
        self._auxiliary_fuel_need()
        if self.user_config["flexestimators"]["filter_fuel_need"]:
            self.activities = self._filter_residual_need(activities=self.activities, index_columns=["unique_id"])
        self.activities = self._cleanup_dataset(activities=self.activities)
        if self.user_config["global"]["write_output_to_disk"]["flex_output"]:
            self.__write_output()
        logging.info("Technical flexibility estimation ended.")
        self.data['activities'] = self.activities
        return self.data




__maintainer__ = "Fabia Miorelli, Eugenio Salavador Arellano Ruiz"
__license__ = "BSD-3-Clause"

import yaml
from pathlib import Path


def read_metadata_config():
    """
    Funtion to load  the needed config for metadata generation.

    Args:
        base_path (Path): _description_

    Returns:
        metadata_config (dict): Dictionary with opened yaml config files
    """
    config_name = "metadata_default"
    config_path = Path((__file__)).parent.parent / "config"
    file_path = (config_path / config_name).with_suffix(".yaml")
    with open(file_path, encoding="utf-8") as ipf:
        metadata_config = yaml.load(ipf, Loader=yaml.SafeLoader)
    return metadata_config


def write_out_metadata(metadata_yaml, file_name):
    with open(file_name, "w", encoding="utf-8") as f:
        yaml.Dumper.ignore_aliases = lambda *args: True
        yaml.dump(metadata_yaml, f, sort_keys=False, allow_unicode=True)



__maintainer__ = "Niklas Wulff, Fabia Miorelli"
__license__ = "BSD-3-Clause"

import logging
import time

from pathlib import Path
import pandas as pd

from ..core.diarybuilders import DiaryBuilder
from ..utils.utils import create_file_name, write_out
from ..utils.metadata import read_metadata_config, write_out_metadata


class ProfileAggregator:
    def __init__(self, configs: dict, data: dict, profiles: DiaryBuilder):
        """
        In the ProfileAggregator, single vehicle profiles are aggregated across all vehicles to gain fleet level
        profiles. Depending on the profile type, different aggregation approaches are used.
        The design pattern is similar as in the diarybuilders. There is one wrapper class, ProfileAggregator, that has
        an instance of the class Aggregator as attribute. This attribute's method perform_aggregation() is then called
        for each profile relevant for the analysis, specifying the profile type as a parameter ('flow' or 'state'). The
        profiles for drain, availability and uncontrolled charging are all flow profiles, whereas the battery level
        profiles are state profiles and thus, are aggregated differently.
        Two options can be given in the user_config for refining profile aggregation: The timespan and the aggregation
        weights. The timespan can be given in the option 'aggregation_timespan', specifying if the resulting profile is
        a daily profile or a weekly profile.
        The flow profiles are aggregated using means or weighted means which can be specified in the profileaggregators
        section of the user_config under the option 'weight_flow_profiles'. The aggregation does not change the temporal
        resolution, it is only related to aggregate the charging profiles from around 100,000 single-vehicle profiles to
        fleet profiles. Thus, after aggregation, there are 5 profiles with the temporal timespan (daily or weekly) and
        the temporal resolution selected in the diary builder before (e.g. 24 values for daily profiles with hourly
        resolution).

        Args:
            configs (dict): A dictionary containing a user_config dictionary and a dev_config dictionary
            activities (pd.DataFrame): A dataframe containing all trip and parking activities
            profiles (DiaryBuilder): An instance of type DiaryBuilder
        """
        self.user_config = configs["user_config"]
        self.dev_config = configs["dev_config"]
        self.dataset = self.user_config["global"]["dataset"]
        self.weighted = self.user_config["profileaggregators"]["weight_flow_profiles"]
        self.data = data
        self.activities = data['activities']
        self.profiles = profiles
        self.drain = profiles.drain
        self.charging_power = profiles.charging_power
        self.uncontrolled_charging = profiles.uncontrolled_charging
        self.max_battery_level = profiles.max_battery_level
        self.min_battery_level = profiles.min_battery_level
        self.drain_weekly = pd.DataFrame()
        self.charging_power_weekly = pd.DataFrame()
        self.uncontrolled_charging_weekly = pd.DataFrame()
        self.max_battery_level_weekly = pd.DataFrame()
        self.min_battery_level_weekly = pd.DataFrame()
        self.aggregator = Aggregator(
            activities=self.activities,
            dataset=self.dataset,
            user_config=self.user_config,
            dev_config=self.dev_config,
            weighted=self.weighted,
        )

    def generate_metadata(self, metadata_config, file_name):
        """
        _summary_

        Args:
            metadata_config (_type_): _description_
            file_name (_type_): _description_

        Returns:
            _type_: _description_
        """
        metadata_config["name"] = file_name
        metadata_config["title"] = (
            "Discetised timeseries with venco.py output profiles at fleet level"
        )
        metadata_config["description"] = "Time discrete profile at fleet level."
        metadata_config["sources"] = [
            f for f in metadata_config["sources"] if f["title"] in self.dataset
        ]
        reference_resource = metadata_config["resources"][0]
        this_resource = reference_resource.copy()
        this_resource["name"] = file_name.rstrip(".csv")
        this_resource["path"] = file_name
        these_fields = [
            f
            for f in reference_resource["schema"][self.dataset]["fields"][
                "profileaggregators"
            ]
        ]
        this_resource["schema"] = {"fields": these_fields}
        metadata_config["resources"].pop()
        metadata_config["resources"].append(this_resource)
        return metadata_config

    def _write_metadata(self, file_name):
        metadata_config = read_metadata_config()
        class_metadata = self.generate_metadata(
            metadata_config=metadata_config, file_name=file_name.name
        )
        write_out_metadata(
            metadata_yaml=class_metadata,
            file_name=file_name.as_posix().replace(".csv", ".metadata.yaml"),
        )

    def aggregate_profiles(self):
        """
        Wrapper method for the ProfileAggregator class. This method is supposed to be called in the overarching venco.py
        workflow.
        """
        if self.user_config["global"]["consider_temperature_cycle_dependency"]["season"] == "all":
            seasons = ["spring", "summer", "winter", "fall"]
            for season in seasons:
                logging.info(f"Aggreagting a weekly profile for season: {season}")
                ids_season = self.activities[self.activities.season == season]["unique_id"].unique()
                self.drain_weekly[season] = self.aggregator.perform_aggregation(profile=self.drain.loc[ids_season], profile_name=f"drain_{season}", method="flow")
                self.charging_power_weekly[season] = self.aggregator.perform_aggregation(profile=self.charging_power.loc[ids_season], profile_name=f"charging_power_{season}", method="flow")
                self.uncontrolled_charging_weekly[season] = self.aggregator.perform_aggregation(profile=self.uncontrolled_charging.loc[ids_season], profile_name=f"uncontrolled_charging_{season}", method="flow")
                self.max_battery_level_weekly[season] = self.aggregator.perform_aggregation(profile=self.max_battery_level.loc[ids_season], profile_name=f"max_battery_level_{season}", method="state")
                self.min_battery_level_weekly[season] = self.aggregator.perform_aggregation(profile=self.min_battery_level.loc[ids_season], profile_name=f"min_battery_level_{season}", method="state")
        else:
            self.drain_weekly = self.aggregator.perform_aggregation(
                profile=self.drain, profile_name="drain", method="flow"
            )
            self.charging_power_weekly = self.aggregator.perform_aggregation(
                profile=self.charging_power, profile_name="charging_power", method="flow"
            )
            self.uncontrolled_charging_weekly = self.aggregator.perform_aggregation(
                profile=self.uncontrolled_charging,
                profile_name="uncontrolled_charging",
                method="flow"
            )
            self.max_battery_level_weekly = self.aggregator.perform_aggregation(
                profile=self.max_battery_level,
                profile_name="max_battery_level",
                method="state"
            )
            self.min_battery_level_weekly = self.aggregator.perform_aggregation(
                profile=self.min_battery_level,
                profile_name="min_battery_level",
                method="state"
            )
        if self.user_config["global"]["write_output_to_disk"]["aggregator_output"]:
            root = Path(self.user_config["global"]["absolute_path"]["vencopy_root"])
            folder = self.dev_config["global"]["relative_path"]["aggregator_output"]
            file_name = create_file_name(
                dev_config=self.dev_config,
                user_config=self.user_config,
                file_name_id="output_profileaggregator"
            )
            self._write_metadata(file_name=root / folder / file_name)
            logging.info("Aggregation finished for all profiles.")


class Aggregator:
    def __init__(
        self,
        activities: pd.DataFrame,
        dataset: str,
        user_config: dict,
        dev_config: dict,
        weighted: bool,
        ):
        """
        Class to perform aggregation of state and flow profiles either on daily or weekly scope.

        Args:
            activities (pd.DataFrame): A dataframe containing all trip and parking activities
            dataset (str): Name of initial mobility data set used for resulting file name annotation
            user_config (dict): A dictionary specifying user-specific options
            dev_config (dict): A dictionary specifying options that are only needed if own development is going on.
            weighted (bool): Shall the aggregation take into account weights from the initial mobility data sets?
        """
        self.dataset = dataset
        self.activities = activities
        self.weighted = weighted
        self.user_config = user_config
        self.dev_config = dev_config
        self.alpha = self.user_config["profileaggregators"]["alpha"]
        self.aggregation_scope = user_config["profileaggregators"][
            "aggregation_timespan"
        ]
        self.weekday_profiles = None

    def _extract_weights(self):
        """
        Get weights from the initial mobility data set stored in the column trip_weight and store it in self.weights.
        """
        self.weights = (
            self.activities.loc[:, ["unique_id", "trip_weight"]]
            .drop_duplicates(subset=["unique_id"])
            .reset_index(drop=True)
            .set_index("unique_id")
        )

    def __basic_aggregation(self):
        """
        Decider function differentiating between daily and weekly scope for the aggregation.
        """
        if self.aggregation_scope == "daily":
            self._aggregate_daily()
        elif self.aggregation_scope == "weekly":
            self._aggregate_weekly()
            self.__compose_week_profile()
        else:
            NotImplementedError(
                "The aggregation timespan can either be daily or weekly."
            )

    def _aggregate_daily(self):
        """
        Function to aggregate all profiles across all days, independently of the day of the week for both, state and
        flow profiles. The result is stored under self.profile_agg.

        Raises:
            NotImplementedError: Raised if this function is called for a state profile and the profile is neither of
            max_battery_level and min_battery_level
        """
        self.daily_profile = pd.DataFrame(
            columns=self.profile.columns, index=range(1, 2)
        )
        cols = ["unique_id", "trip_weight"]
        self.activities_subset = (
            self.activities[cols]
            .copy()
            .drop_duplicates(subset=["unique_id"])
            .reset_index(drop=True)
        )
        self.activities_weekday = pd.merge(
            self.profile, self.activities_subset, on="unique_id", how="inner"
        )
        self.activities_weekday = self.activities_weekday.set_index("unique_id")
        if self.method == "flow":
            if self.weighted:
                # weekday_subset = weekday_subset.drop("trip_start_weekday", axis=1)
                # aggregate activities_weekday to one profile by multiplying by weights
                weight_sum = sum(self.activities_weekday.trip_weight)
                daily_subset_weight = self.activities_weekday.apply(
                    lambda x: x * self.activities_weekday.trip_weight.values
                )
                daily_subset_weight = daily_subset_weight.drop("trip_weight", axis=1)
                daily_subset_weight_agg = daily_subset_weight.sum() / weight_sum
                self.weekday_profiles = daily_subset_weight_agg
            else:
                daily_subset = self.activities_weekday.drop(
                    columns=["trip_weight"], axis=1
                ).reset_index(drop=True)
                self.weekday_profiles = daily_subset.mean(axis=0)
        elif self.method == "state":
            daily_subset = self.activities_weekday.drop(
                columns=["trip_weight"]
            ).reset_index(drop=True)
            daily_subset = daily_subset.convert_dtypes()
            if self.profile_name == "max_battery_level":
                self.weekday_profiles = daily_subset.quantile(1 - (self.alpha / 100))
            elif self.profile_name == "min_battery_level":
                self.weekday_profiles = daily_subset.quantile(self.alpha / 100)
            else:
                raise NotImplementedError(
                    f"An unknown profile {self.profile_name} was selected."
                )

    def _aggregate_weekly(self, by_column: str = "trip_start_weekday"):
        """
        Function to aggregate all profiles depending on the day of the week to be able to create a weekly profile from
        individual daily aggregates. Two differentiations are made. Firstly, the function differentiates between flow
        and state profiles (stored in self.method), secondly if weights should be applied to the daily aggregations
        stored in self.weighted.

        Args:
            by_column (str, optional): Column to be used to perform the aggregation. Defaults to "trip_start_weekday".
        """

        self.weekday_profiles = pd.DataFrame(
            columns=self.profile.columns, index=range(1, 8)
        )
        cols = ["unique_id", "trip_weight"] + [by_column]
        self.activities_subset = (
            self.activities[cols]
            .copy()
            .drop_duplicates(subset=["unique_id"])
            .reset_index(drop=True)
        )
        self.activities_weekday = pd.merge(
            self.profile, self.activities_subset, on="unique_id", how="inner"
        )
        # self.profile.drop('unique_id', axis=1, inplace=True)
        self.activities_weekday = self.activities_weekday.set_index("unique_id")
        # Compose weekly profile from 7 separate profiles
        if self.method == "flow":
            if self.weighted:
                self.__calculate_weighted_mean_flow_profiles(
                    by_column="trip_start_weekday"
                )
            else:
                self.__calculate_average_flow_profiles(by_column="trip_start_weekday")
        elif self.method == "state":
            self.__aggregate_state_profiles(
                by_column="trip_start_weekday", alpha=self.alpha
            )

    def __calculate_average_flow_profiles(self, by_column: str):
        """
        Iterates through all unique elements given in by_column and aggregates flow profiles for the set of unique_ids
        that are associated with each element of the currently iterated instance. The resulting profile is stored in
        self.pofile_agg.

        Args:
            by_column (str): The column to split the profile unique_ids by, defaulting to trip_start_weekday
        """
        for idate in self.activities_weekday[by_column].unique():
            weekday_subset = self.activities_weekday[
                self.activities_weekday[by_column] == idate
            ].reset_index(drop=True)
            weekday_subset = weekday_subset.drop(
                columns=["trip_start_weekday", "trip_weight"], axis=1
            )
            weekday_subset_agg = weekday_subset.sum(axis=0)
            self.weekday_profiles.iloc[idate - 1] = weekday_subset_agg

    def __calculate_weighted_mean_flow_profiles(self, by_column: str):
        """
        The same as self.__calculate_average_flow_profiles(), but applying weighted means using weights from the initial
        mobility data set.

        Args:
            by_column (str): The column to split the profile unique_ids by, defaulting to trip_start_weekday
        """
        for idate in self.activities_weekday[by_column].unique():
            weekday_subset = self.activities_weekday[
                self.activities_weekday[by_column] == idate
            ].reset_index(drop=True)
            weekday_subset = weekday_subset.drop("trip_start_weekday", axis=1)
            # aggregate activities_weekday to one profile by multiplying by weights

            # DEPRECATED
            # weight_sum = sum(weekday_subset.trip_weight)

            weekday_subset_weight = weekday_subset.apply(
                lambda x: x * weekday_subset.trip_weight.values
            )
            weekday_subset_weight = weekday_subset_weight.drop("trip_weight", axis=1)
            weekday_subset_weight_agg = (
                weekday_subset_weight.sum()
            )  # / weight_sum DEPRECATED
            self.weekday_profiles.iloc[idate - 1] = weekday_subset_weight_agg

    def __aggregate_state_profiles(self, by_column: str, alpha: int = 10):
        """
        Selects the alpha (100 - alpha) percentile from maximum battery level (minimum batttery level) profile for each
        hour. If alpha = 10, the 10%-biggest (10%-smallest) value is selected, all values beyond are disregarded as
        outliers.

        Args:
            by_column (str): The column to split the profile unique_ids by, defaulting to trip_start_weekday
            alpha (int, optional): Percentage, giving the amount of profiles whose mobility demand can not be fulfilled
                after selection. Defaults to 10.

        Raises:
            NotImplementedError: This error is raised if self.profile_name is not one of max_battery_level and
            min_battery_level
        """
        for idate in self.activities_weekday[by_column].unique():
            levels = self.activities_weekday.copy()
            weekday_subset = levels[levels[by_column] == idate].reset_index(drop=True)
            weekday_subset = weekday_subset.drop(
                columns=["trip_start_weekday", "trip_weight"]
            )
            weekday_subset = weekday_subset.convert_dtypes()
            if self.profile_name.startswith("max_battery_level"):
                self.weekday_profiles.iloc[idate - 1] = weekday_subset.quantile(
                    alpha / 100
                )
            elif self.profile_name.startswith("min_battery_level"):
                self.weekday_profiles.iloc[idate - 1] = weekday_subset.quantile(
                    1 - (alpha / 100)
                )
            else:
                raise NotImplementedError(
                    f"An unknown profile {self.profile_name} was selected."
                )
                
    def __compose_week_profile(self):
        """
        Input is self.weekday_profiles. Method only works if aggregation is weekly. Check if any day of the week is not
        filled, copy line above in that case
        """
        if self.weekday_profiles.isna().any(axis=1).any():
            index_empty_rows = (
                self.weekday_profiles[self.weekday_profiles.isna().any(axis=1)].index
                - 1
            )
            for empty_row in index_empty_rows:
                if empty_row == 5 or empty_row == 7:
                    self.weekday_profiles.iloc[empty_row] = self.weekday_profiles.iloc[
                        empty_row - 1
                    ]
                else:
                    self.weekday_profiles.iloc[empty_row] = self.weekday_profiles.iloc[
                        empty_row + 1
                    ]
        
        self.weekday_profiles.index.name = "weekday"
        self.weekday_profiles = self.weekday_profiles.stack().unstack(0)
        if self.weekday_profiles.empty:
            raise ValueError(r"The 'weekday_profiles' DataFrame seems to be empty for one or more seasons. Please add more columns to debug mode.")
        self.weekday_profiles = pd.concat(
            [
                self.weekday_profiles[1],
                self.weekday_profiles[2],
                self.weekday_profiles[3],
                self.weekday_profiles[4],
                self.weekday_profiles[5],
                self.weekday_profiles[6],
                self.weekday_profiles[7],
            ],
            ignore_index=True,
        )
        self.weekday_profiles = pd.DataFrame(self.weekday_profiles, columns=[self.profile_name])
        logging.info(f"Finished generating a weekly profile for the vehicle segment {self.user_config['dataparsers']['vehicle_segment'][self.dataset]} with {self.profile_name}.")

    def __write_output(self):
        """
        Write out the profile given in profile adding the self.profile_name and the string "output_profileaggregator"
        (overarching file label) to the written filename.
        """
        if self.user_config["global"]["write_output_to_disk"]["aggregator_output"]:
            root = Path(self.user_config["global"]["absolute_path"]["vencopy_root"])
            folder = self.dev_config["global"]["relative_path"]["aggregator_output"]
            self.dev_config["global"]["additional_label"] = "_" + self.profile_name + "_"
            file_name = create_file_name(
                dev_config=self.dev_config,
                user_config=self.user_config,
                file_name_id="output_profileaggregator",
                dataset=self.dataset,
            )
            write_out(data=self.weekday_profiles, path=root / folder / file_name)

    def perform_aggregation(
        self, profile: pd.DataFrame, profile_name: str, method: str
    ) -> pd.DataFrame:
        """
        Inner wrapper function (of sub-class) that extracts weights, performs the aggregation and writes the output to
        disk. Internal class attributes are reset to None after write-out.

        Args:
            profile (pd.DataFrame): Profiles to be aggregated
            profile_name (str): Descriptor used for filename annotation
            method (str): Can be one of "state" or "flow"

        Returns:
            pd.DataFrame: Returns the aggregated profile
        """
        self.profile = profile
        self.profile_name = profile_name
        self.method = method
        logging.info(
            f"Starting to aggregate {self.profile_name} to fleet level based on day of the week."
        )
        start_time_agg = time.time()
        if self.weighted:
            self._extract_weights()
        self.__basic_aggregation()
        if self.user_config["global"]["write_output_to_disk"]["aggregator_output"]:
            self.__write_output()
        logging.info(f"Aggregation finished for {self.profile_name}.")
        elapsed_time_agg = time.time() - start_time_agg
        logging.info(f"Needed time to aggregate {self.profile_name}: {elapsed_time_agg}.")
        self.profile = None
        self.profile_name = None
        self.method = None
        return self.weekday_profiles



__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"

import logging
from pathlib import Path

import pandas as pd
from ...core.dataparsers.dataparsers import IntermediateParsing
from ...core.dataparsers.parkinference import ParkInference


class ParseKiD(IntermediateParsing):
    def __init__(self, configs: dict, dataset: str):
        """
        Inherited data class to differentiate between abstract interfaces such
        as vencopy internal variable namings and data set specific functions
        such as filters. Specific class for the German KiD dataset.

        Args:
            configs (dict): A dictionary containing a user_config dictionary and a dev_config dictionary.
            dataset (str): Abbreviation of the National Travel Survey to be parsed.
        """
        super().__init__(configs=configs, dataset=dataset)
        self.park_inference = ParkInference(configs=configs)

    def _load_unencrypted_data(self):
        """
        Loads the dataset specified in the user_config.
        """
        raw_data_path_trips = (
            Path(self.user_config["global"]["absolute_path"][self.dataset])
            / self.dev_config["global"]["files"][self.dataset]["trips_data_raw"]
        )
        raw_data_path_vehicles = (
            Path(self.user_config["global"]["absolute_path"][self.dataset])
            / self.dev_config["global"]["files"][self.dataset]["vehicles_data_raw"]
        )
        raw_data_trips = pd.read_stata(
            raw_data_path_trips,
            convert_categoricals=False,
            convert_dates=False,
            preserve_dtypes=False,
        )
        raw_data_vehicles = pd.read_stata(
            raw_data_path_vehicles,
            convert_categoricals=False,
            convert_dates=False,
            preserve_dtypes=False,
        )
        raw_data_vehicles.set_index("k00", inplace=True)
        raw_data = raw_data_trips.join(raw_data_vehicles, on="k00")
        self.raw_data = raw_data
        logging.info(f"Finished loading {len(self.raw_data)} " f"rows of raw data of type .dta.")

    @staticmethod
    def _change_separator(trips):
        """
        Replaces commas with dots in the dataset (German datasets).

        Args:
            trips (pd.DataFrame): A dataframe containing all trips.

        Returns:
            pd.DataFrame: A dataframe containing all trips.
        """
        for i, x in enumerate(list(trips.trip_distance)):
            trips.at[i, "trip_distance"] = x.replace(",", ".")
        for i, x in enumerate(list(trips.trip_weight)):
            trips.at[i, "trip_weight"] = x.replace(",", ".")
        return trips

    def __add_string_columns(self, weekday=True, purpose=True, purpose_from=True, vehicle_segment=True):
        """
        Adds string columns for either weekday or purpose.

        Args:
            weekday (bool, optional): Boolean identifier if weekday should be added in a separate column as string. Defaults to True.
            purpose (bool, optional): Boolean identifier if purpose should be added in a separate column as string. Defaults to True.
            vehicle_segment (bool, optional): Boolean identifier if the vehicle segment should be added in a separate column as string. Defaults to True.
        """
        if weekday:
            self._add_string_column_from_variable(col_name="weekday_string", var_name="trip_start_weekday")
        if purpose:
            self._add_string_column_from_variable(col_name="purpose_string", var_name="trip_purpose")
        if purpose_from:
            self._add_string_column_from_variable(col_name="purpose_from_string", var_name="trip_purpose_from")
        if vehicle_segment:
            self._add_string_column_from_variable(col_name="vehicle_segment_string", var_name="vehicle_segment")

    @staticmethod
    def _extract_timestamps(trips):
        """
        Extracts timestamps from the trip start date.

        Args:
            trips (pd.DataFrame): A dataframe containing all trips.

        Returns:
            pd.DataFrame: A dataframe containing all trips.
        """
        trips["trip_start_date"] = pd.to_datetime(trips["trip_start_date"], format="%d.%m.%Y")
        trips["trip_start_year"] = trips["trip_start_date"].dt.year
        trips["trip_start_month"] = trips["trip_start_date"].dt.month
        trips["trip_start_day"] = trips["trip_start_date"].dt.day
        trips["trip_start_weekday"] = trips["trip_start_date"].dt.weekday
        trips["trip_start_week"] = trips["trip_start_date"].dt.isocalendar().week
        trips["trip_start_week"] = trips["trip_start_week"].astype(int)
        trips["trip_start_hour"] = pd.to_datetime(trips["trip_start_clock"], format="%H:%M").dt.hour
        trips["trip_start_minute"] = pd.to_datetime(trips["trip_start_clock"], format="%H:%M").dt.minute
        trips["trip_end_hour"] = pd.to_datetime(trips["trip_end_clock"], format="%H:%M").dt.hour
        trips["trip_end_minute"] = pd.to_datetime(trips["trip_end_clock"], format="%H:%M").dt.minute
        return trips

    @staticmethod
    def _update_end_timestamp(trips):
        """
        Separate implementation for the KID dataset. Overwrites parent method.

        Args:
            trips (pd.DataFrame): A dataframe containing all trips.

        Returns:
            pd.DataFrame: A dataframe containing all trips.
        """
        trips["trip_end_next_day"] = False
        trips["trip_end_next_day"] = trips["trip_end_next_day"].where(~(trips["timestamp_start"] > trips["timestamp_end"]), True)
        ends_following_day = trips["trip_end_next_day"] == True 
        trips.loc[ends_following_day, "timestamp_end"] = trips.loc[
            ends_following_day, "timestamp_end"
        ] + pd.offsets.Day(1)
        return trips

    @staticmethod
    def _exclude_hours(trips):
        """
        Removes trips where either start and end trip time are missing. KID-specific function.

        Args:
            trips (pd.DataFrame): A dataframe containing all trips.

        Returns:
            pd.DataFrame: A dataframe containing all trips.
        """
        trips.drop(trips.loc[trips["trip_start_clock"].str.contains("-1")].index, inplace=True)
        trips.drop(trips.loc[trips["trip_end_clock"].str.contains("-1")].index, inplace=True)
        return trips

    @staticmethod
    def _cleanup_dataset(activities):
        activities.drop(
            columns=['vehicle_id',
                     'vehicle_segment',
                     'vehicle_segment_string',
                     'trip_start_date',
                     'trip_start_day',
                     'trip_scale_factor',
                     'trip_end_next_day',
                     'trip_is_intermodal',
                     'trip_purpose',
                     'weekday_string',
                     'is_first_trip'], inplace=True)
        return activities

    def process(self) -> pd.DataFrame:
        """
        Wrapper function for harmonising and filtering the trips dataset as well
        as adding parking rows.
        """
        self._load_data()
        self._select_columns()
        self._harmonise_variables()
        self._harmonise_variables_unique_id_names()
        self._change_separator(trips=self.trips)
        self._convert_types()
        self.trips = self._exclude_hours(trips=self.trips)
        self._extract_timestamps(trips=self.trips)
        self.__add_string_columns()
        self._compose_start_and_end_timestamps()
        self._update_end_timestamp(trips=self.trips)
        self._assign_season(self.trips)
        if self.user_config["dataparsers"]["seasons"]["filter_season_post_assignment"]:
            self.filter_for_season(self.trips)
        self._check_filter_dict(dictionary=self.filters)
        self._filter(filters=self.filters)
        self._filter_consistent_hours(dataset=self.trips)
        self._extract_df_debug_mode()
        self.data = self.park_inference.add_parking_rows(trips=self.trips)
        self._subset_vehicle_segment()
        self._subset_area_type()
        self.data['activities'] = self._cleanup_dataset(activities=self.data['activities'])
        self.write_output()
        logging.info("Parsing KiD dataset completed.")
        return self.data



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import sphinx_rtd_theme
sys.path.insert(0, os.path.abspath('.'))
sys.path.append(os.path.abspath('..'))
sys.path.append(os.path.abspath('../utils'))
sys.path.append(os.path.abspath('../core'))

# -- Project information -----------------------------------------------------

project = 'vencopy'
copyright = '2020-2024, German Aerospace Center (DLR)'
author = 'German Aerospace Center (DLR)'

# The full version, including alpha/beta/rc tags
release = 'November 2024'

highlight_language = 'python3'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx_rtd_theme',
              'sphinx.ext.napoleon',
              'sphinx.ext.autosectionlabel',
              'rst2pdf.pdfbuilder']

pdf_documents = [('index', 'vencopy', 'vencopypdf'),]

source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'markdown',
    '.md': 'markdown',
}
# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
html_theme_path = ['_themes']
html_theme_options = {
    # 'style_nav_header_background': 'blue',
    # Toc options
    'collapse_navigation': True,
    'sticky_navigation': False,
    'navigation_depth': 3,
    'includehidden': True,
    'titles_only': False
}
# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False
# html_css_files = ['style.css']
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


__maintainer__ = "Fabia Miorelli"
__license__ = "BSD-3-Clause"

import pytest

import pandas as pd

from ...vencopy.core.diarybuilders import TimeDiscretiser

# NOT TESTED: __write_output(), discretise()


@pytest.fixture
def sample_configs():
    configs = {
        'user_config': {
            'global': {
                'dataset': "dataset1",
                'debug': False,
                'absolute_path': {
                    'dataset1': '/path/to/dataset1',
                    'dataset2': '/path/to/dataset2'
                    }
                },
            'diarybuilders': {
                'time_resolution': 15
            }
            },
        'dev_config': {
            'dataparsers': {
                'data_variables': False
                }
            }
        }
    return configs


@pytest.fixture
def sample_activities():
    activities = pd.DataFrame({
        "activity_id": [1, 2, 3, 4],
        "activity_duration": [pd.Timedelta(minutes=76), pd.Timedelta(minutes=80), pd.Timedelta(0), pd.Timedelta(minutes=45)],
        "timestamp_start": pd.DatetimeIndex(["2023-09-12 08:00:00", "2023-09-12 10:30:00", "2023-09-12 10:30:00", "2023-09-12 10:00:00"]),
        "timestamp_end": pd.DatetimeIndex(["2023-09-12 09:16:00", "2023-09-12 11:50:00", "2023-09-12 10:30:00", "2023-09-12 10:45:00"])
        })
    return activities


def test_timediscretiser_init(sample_configs,):

    discretiser = TimeDiscretiser(dataset="dataset1", user_config=sample_configs["user_config"], dev_config=sample_configs["dev_config"], time_resolution=15)

    assert discretiser.activities == None
    assert discretiser.dev_config == sample_configs["dev_config"]
    assert discretiser.user_config == sample_configs["user_config"]
    assert discretiser.dataset == "dataset1"
    assert discretiser.time_resolution == 15
    assert discretiser.quantum == pd.Timedelta(value=1, unit="min")
    assert discretiser.is_week == False
    assert discretiser.data_to_discretise is None



__maintainer__ = "Niklas Wulff, Fabia Miorelli"
__license__ = "BSD-3-Clause"

import os
import yaml
import shutil
import pathlib
import click
import vencopy


@click.command()
@click.option(
    "--name",
    default="vencopy_user",
    prompt="Please type the user folder name:",
    help="The folder name of the venco.py user folder created at command line working directory",
)
# Tutorials will be updated on next release
@click.option(
    "--tutorials",
    default="false",
    help="Specify if tutorials should be copied to the user folder on set up." "Defaults to true",
)
def create(name: str, tutorials: bool):
    """
    Folder set up after installation.


    Args:
        name (str): _description_
        tutorials (bool): _description_
    """
    cwd = pathlib.Path(os.getcwd())
    target = cwd / name
    source = pathlib.Path(vencopy.__file__).parent.resolve()
    if not os.path.exists(target):
        os.mkdir(target)
        setup_folders(src=source, trg=target, tutorials=tutorials)
        click.echo(f"venco.py user folder created under {target}")
    elif os.path.exists(target) and not os.path.exists(target / "run.py"):
        setup_folders(src=source, trg=target, tutorials=tutorials)
        click.echo(f"venco.py user folder filled under {target}")
    else:
        click.echo(
            "File run.py already exists in specified folder, for a new setup please specify a non-existent "
            "folder name"
        )


def setup_folders(src: pathlib.Path, trg: pathlib.Path, tutorials: bool):
    """
    Setup function to create a vencopy user folder and to copy run, config and tutorial files from the package source.

    Args:
        src (pathlib.Path): Absolute path to the vencopy package source folder
        trg (pathlib.Path): Absolute path to the vencopy user folder
        tutorials (bool): Boolean, if true (default) tutorials are being copied from package source to user folder
    """
    os.mkdir(trg / "config")
    os.mkdir(trg / "core")
    os.mkdir(trg / "output")
    os.mkdir(trg / "output" / "dataparser")
    os.mkdir(trg / "output" / "diarybuilder")
    os.mkdir(trg / "output" / "gridmodeller")
    os.mkdir(trg / "output" / "flexestimator")
    os.mkdir(trg / "output" / "profileaggregator")
    os.mkdir(trg / "output" / "postprocessor")
    os.mkdir(trg / "utils")
    shutil.copy(src=src / "run.py", dst=trg)
    shutil.copytree(src=src / "config", dst=trg / "config")
    if tutorials:
        raise (NotImplementedError("Tutorials for the new venco.py iteration are not yet implemented."))
        # shutil.copytree(src=src / "tutorials", dst=trg / "tutorials")
    update_config(new_vencopy_root=trg)


def update_config(new_vencopy_root: pathlib.Path):
    """
    _missing_

    Args:
        new_vencopy_root (pathlib.Path): _description_
    """
    with open(new_vencopy_root / "config" / "user.yaml") as f:
        user_cfg = yaml.load(f, Loader=yaml.SafeLoader)

    user_cfg["absolute_path"]["vencopy_root"] = new_vencopy_root.__str__()

    with open(new_vencopy_root / "config" / "user.yaml", "w") as f:
        yaml.dump(user_cfg, f)


if __name__ == "__main__":
    create()



__maintainer__ = "Niklas Wulff, Fabia Miorelli"
__license__ = "BSD-3-Clause"

import logging
import os
import yaml

from pathlib import Path

import pandas as pd

# Constants that define Seasons
winter_start = pd.to_datetime("12-21", format="%m-%d").dayofyear
spring_start = pd.to_datetime("03-20", format="%m-%d").dayofyear
summer_start = pd.to_datetime("06-21", format="%m-%d").dayofyear
fall_start = pd.to_datetime("09-23", format="%m-%d").dayofyear


def load_configs(base_path: Path) -> dict:
    """
    Generic function to load and open yaml config files.
    Uses pathlib syntax for windows, max, linux compatibility,
    see https://realpython.com/python-pathlib/ for an introduction.

    Args:
        base_path (Path): _description_

    Returns:
        configs (dict): Dictionary with opened yaml config files
    """
    config_names = ("user_config", "dev_config")
    config_path = Path(base_path) / "config"
    configs = {}
    for config_name in config_names:
        file_path = (config_path / config_name).with_suffix(".yaml")
        with open(file_path) as ipf:
            configs[config_name] = yaml.load(ipf, Loader=yaml.SafeLoader)
    return configs


def return_lowest_level_dict_keys(dictionary: dict, lst: list = None) -> list:
    """
    Returns the lowest level keys of dictionary and returns all of them
    as a list. The parameter lst is used as
    interface between recursion levels.

    Args:
        dictionary (dict): Dictionary of variables
        lst (list, optional): List used as interface between recursion levels. Defaults to None.

    Returns:
        list: list with all the bottom level dictionary keys
    """
    if lst is None:
        lst = []
    for i_key, i_value in dictionary.items():
        if isinstance(i_value, dict):
            lst = return_lowest_level_dict_keys(i_value, lst)
        elif i_value is not None:
            lst.append(i_key)
    return lst


def return_lowest_level_dict_values(dictionary: dict, lst: list = None) -> list:
    """
    Returns a list of all dictionary values of the last dictionary level
    (the bottom) of dictionary. The parameter
    lst is used as an interface between recursion levels.

    Args:
        dictionary (dict): Dictionary of variables
        lst (list, optional): List used as interface to next recursion. Defaults to None.

    Returns:
        list: List with all the bottom dictionary values
    """
    if lst is None:
        lst = []
    for _, i_value in dictionary.items():
        if isinstance(i_value, dict):
            lst = return_lowest_level_dict_values(i_value, lst)
        elif i_value is not None:
            lst.append(i_value)
    return lst


def replace_vec(
    series, year=None, month=None, day=None, hour=None, minute=None, second=None
) -> pd.Series:
    """
    _summary_

    Args:
        series (_type_): _description_
        year (_type_, optional): _description_. Defaults to None.
        month (_type_, optional): _description_. Defaults to None.
        day (_type_, optional): _description_. Defaults to None.
        hour (_type_, optional): _description_. Defaults to None.
        minute (_type_, optional): _description_. Defaults to None.
        second (_type_, optional): _description_. Defaults to None.

    Returns:
        pd.Series: _description_
    """
    replacement = pd.to_datetime(
        {
            "year": (
                series.dt.year if year is None else [year for i in range(len(series))]
            ),
            "month": (
                series.dt.month
                if month is None
                else [month for i in range(len(series))]
            ),
            "day": series.dt.day if day is None else [day for i in range(len(series))],
            "hour": (
                series.dt.hour if hour is None else [hour for i in range(len(series))]
            ),
            "minute": (
                series.dt.minute
                if minute is None
                else [minute for i in range(len(series))]
            ),
            "second": (
                series.dt.second
                if second is None
                else [second for i in range(len(series))]
            ),
        }
    )
    return replacement


def create_output_folders(configs: dict):
    """
    Function to crete vencopy output folder and subfolders

    Args:
        configs (dict): _description_
    """
    root = Path(configs["user_config"]["global"]["absolute_path"]["vencopy_root"])
    main_dir = "output"
    if not os.path.exists(Path(root / main_dir)):
        os.mkdir(Path(root / main_dir))
    sub_dirs = (
        "dataparser",
        "diarybuilder",
        "gridmodeller",
        "flexestimator",
        "profileaggregator",
        "postprocessor",
    )
    for sub_dir in sub_dirs:
        if not os.path.exists(Path(root / main_dir / sub_dir)):
            os.mkdir(Path(root / main_dir / sub_dir))


def create_file_name(dev_config: dict, user_config: dict, file_name_id: str, dataset: str, suffix: str = "csv") -> str:
    """
    Generic method used for fileString compilation throughout the venco.py framework. This method does not write any
    files but just creates the file name including the filetype suffix.

    Args:
        dev_config (dict): _description_
        user_config (dict): _description_
        file_name_id (str): ID of respective data file as specified in global config
        dataset (str): Dataset
        manual_label (str, optional):  Optional manual label to add to file_name. Defaults to "".
        suffix (str, optional): _description_. Defaults to "csv".

    Returns:
        str: Full name of file to be written.
    """
    additional_label = dev_config['global']['additional_label']
    run_label = user_config['global']['run_label']
    if len(run_label) == 0:
        return f"{dev_config['global']['disk_file_names'][file_name_id]}_{dataset}.{suffix}"
    if dataset is None:
        return f"{dev_config['global']['disk_file_names'][file_name_id]}_{additional_label}_{run_label}.{suffix}"
    if len(additional_label) == 0:
        return f"{dev_config['global']['disk_file_names'][file_name_id]}_{dataset}_{run_label}.{suffix}"
    return f"{dev_config['global']['disk_file_names'][file_name_id]}_{additional_label}_{dataset}_{run_label}.{suffix}"


def write_out(data: pd.DataFrame, path: Path):
    """
    Utility function to write the DataFrame given in data to the location given in path.

    Args:
        data (pd.DataFrame): Any DataFrame to write to disk
        path (Path): Location on the disk
    """
    data.to_csv(path, header=True)
    logging.info(f"Dataset written to {path}.")


def calculate_shortest_hourly_distance(start_hour, end_hour):
    """
    Function to calculate the shortest hourly distance between to hours
    within a 24 hour cycle

    Args:
        start_hour (float): First hour value
        end_hour (float): Second hour value

    Returns:
        float: Distance in hours
    """
    # Calculate the absolute difference between the hours
    hour_difference = abs(end_hour - start_hour)

    # Calculate the shortest distance considering a 24-hour cycle
    shortest_distance = min(hour_difference, 24 - hour_difference)

    return shortest_distance


def calculate_longest_hourly_distance(start_hour, end_hour):
    """
    Function to calculate the longest hourly distance between to hours
    within a 24 hour cycle

    Args:
        start_hour (float): First hour value
        end_hour (float): Second hour value

    Returns:
        float: Distance in hours
    """
    # Calculate the absolute difference between the hours
    hour_difference = abs(end_hour - start_hour)

    # Calculate the longest distance considering a 24-hour cycle
    longest_distance = max(hour_difference, 24 - hour_difference)

    return longest_distance


def adjust_factor_range_by_temperature_range(user_config, category, season):
    """
    This function calculates the changing factor of the seasonal factor
    based to the Hour_of_the_day.
    The season with the maximum temperature range within a day is used to
    be a reference between min and may consumption
    Seasons with lower temperature ranges are adapted to the maximum range
    (based on the logic that consumption factors change more within a day
    if the temperature range is very high)
    Args:
        season (string): The season of the activitiy

    Returns:
        season_min_factor: The factor for the season at the minimum
        consumption hour of a day
        season_max_factor: The factor for the season at the maximum
        consumption hour of a day
        season_neutral_factor: The factor for the season the hour
        between min and max
    """
    seasons_max_range = max(
        user_config[
            "global"][
            "consider_temperature_cycle_dependency"][
            "day_night_factors"][
            season][
            'temperature_range'] for season in ["winter",
                                                "spring",
                                                "summer",
                                                "fall"]
        )
    season_range = user_config[
        "global"][
        "consider_temperature_cycle_dependency"][
        "day_night_factors"][
        season][
        "temperature_range"]
    season_neutral_factor = category[season+"_factor"]
    season_min_factor = season_neutral_factor * (
        1 - season_range / seasons_max_range)
    season_max_factor = season_neutral_factor * (
        1 + season_range / seasons_max_range)

    return season_min_factor, season_max_factor, season_neutral_factor


def calculate_daily_seasonal_factor(user_config,
                                    season,
                                    category,
                                    hour_of_trip):
    """
    This function calulates the hours of neutral consumption as well as
    the range between the extremes and neutrals
    This is then used to calculate weigthing of the daily factors

    Args:
        season (string): The season of the activitiy
        hour_of_trip (double): Hour of the trip based on data timestamp

    Returns:
        factor: daily consumption factor that is added to seasonal
        consumption
        """

    season_min_factor, season_max_factor, season_neutral_factor = adjust_factor_range_by_temperature_range(user_config, category, season)

    max_hour = user_config["global"]["consider_temperature_cycle_dependency"][
        "day_night_factors"][season]["highest_temperature_hour"]
    min_hour = user_config["global"]["consider_temperature_cycle_dependency"][
        "day_night_factors"][season]["lowest_temperature_hour"]

    range_within = calculate_shortest_hourly_distance(max_hour, min_hour)
    range_outside = calculate_longest_hourly_distance(max_hour, min_hour)

    hour_range_min_hod = float(calculate_shortest_hourly_distance(
        hour_of_trip, min(max_hour, min_hour)))
    hour_range_max_hod = float(calculate_shortest_hourly_distance(
        hour_of_trip, max(max_hour, min_hour)))

    if hour_of_trip >= max(max_hour, min_hour):
        if hour_range_max_hod < hour_range_min_hod:
            if user_config["global"]["consider_temperature_cycle_dependency"][
                "day_night_factors"][season][
                    'consumption_at_highest_hour'] == 'max':
                max_factor = season_max_factor
                factor = max_factor - (
                    abs(max_factor - season_min_factor)
                    / range_outside
                    ) * hour_range_max_hod
            else:
                max_factor = season_min_factor
                factor = max_factor + (
                    abs(max_factor - season_max_factor)
                    / range_outside
                    ) * hour_range_max_hod
        else:
            if user_config["global"]["consider_temperature_cycle_dependency"][
                "day_night_factors"][season][
                    'consumption_at_highest_hour'] == 'max':
                min_factor = season_max_factor
                factor = min_factor - (
                    abs(min_factor - season_min_factor)
                    / range_outside
                    ) * hour_range_min_hod
            else:
                min_factor = season_min_factor
                factor = min_factor + (
                    abs(min_factor - season_min_factor)
                    / range_outside
                    ) * hour_range_min_hod

    elif hour_of_trip <= min_hour:

        if hour_range_max_hod < hour_range_min_hod:
            if user_config["global"]["consider_temperature_cycle_dependency"][
                "day_night_factors"][season][
                    'consumption_at_highest_hour'] == "max":
                max_factor = season_max_factor
                factor = max_factor - (
                    abs(max_factor - season_min_factor)
                    / range_outside
                    ) * hour_range_max_hod
            else:
                max_factor = season_min_factor
                factor = max_factor + (
                    abs(max_factor - season_max_factor)
                    / range_outside
                    ) * hour_range_max_hod
        else:
            if user_config["global"]["consider_temperature_cycle_dependency"][
                "day_night_factors"][season][
                    'consumption_at_highest_hour'] == "max":
                min_factor = season_max_factor
                factor = min_factor - (
                    abs(min_factor - season_min_factor)
                    / range_outside
                    ) * hour_range_min_hod
            else:
                min_factor = season_min_factor
                factor = min_factor + (
                    abs(min_factor - season_max_factor)
                    / range_outside
                    ) * hour_range_min_hod
    else:
        if hour_range_max_hod < hour_range_min_hod:
            if user_config["global"]["consider_temperature_cycle_dependency"][
                "day_night_factors"][season][
                    'consumption_at_highest_hour'] == "max":
                max_factor = season_max_factor
                factor = max_factor - (
                    abs(max_factor - season_min_factor)
                    / range_within
                    ) * hour_range_max_hod
            else:
                max_factor = season_min_factor
                factor = max_factor + (
                    abs(max_factor - season_max_factor)
                    / range_within
                    ) * hour_range_max_hod

        elif hour_range_max_hod > hour_range_min_hod:
            if user_config["global"]["consider_temperature_cycle_dependency"][
                "day_night_factors"][season][
                    'consumption_at_highest_hour'] == "max":
                min_factor = season_max_factor
                factor = min_factor - (
                    abs(min_factor - season_min_factor)
                    / range_within
                    ) * hour_range_max_hod
            else:
                min_factor = season_min_factor
                factor = min_factor + (
                    abs(min_factor - season_max_factor)
                    / range_within
                    ) * hour_range_max_hod
        else:
            factor = season_neutral_factor

    return factor



__maintainer__ = "Niklas Wulff, Fabia Miorelli"
__license__ = "BSD-3-Clause"

import logging

import pandas as pd
import numpy as np
import datetime

from ...utils.utils import replace_vec


class ParkInference:
    def __init__(self, configs: dict):
        """
        Class that provides functionality to transform trip chains from national travel surveys to activity chains
        comprising both, trip and park activities.

        Args:
            configs (dict): A dictionary containing a user_config dictionary and a dev_config dictionary
        """
        self.user_config = configs["user_config"]
        self.dev_config = configs["dev_config"]
        self.dataset = configs["user_config"]["global"]["dataset"]
        self.season = self.user_config["global"]["consider_temperature_cycle_dependency"]["season"]
        self.activities_raw = None
        self.immobile_vehicles = None
        self.vehicle_numbers = []
        self.seasonal_vehicle_numbers = pd.DataFrame()
        self.seasons = ["winter", "spring", "summer", "fall"]
        self.overnight_splitter = OvernightSplitter()
        self.data = {}

    def add_parking_rows(self, trips: pd.DataFrame) -> pd.DataFrame:
        """
        Wrapper function generating park activity rows between the trip data from the original MID dataset. Some
        utility attributes are being added such as is_first_activity, is_last_activity or the unique_id of the next and
        previous activity. Redundant time observations are dropped after timestamp creation for start and end time of
        each activity. Overnight trips (e.g. extending from 23:00 at survey day to 1:30 on the consecutive day) are
        split up into two trips. The first one extends to the end of the day (00:00) and the other one is appended
        to the activity list before the first parking activity (0:00-1:30). The trip distance is split between the two
        based on the time.

        Args:
            trips (pd.DataFrame): Trip chain data set from a travel survey.

        Returns:
            self.activities_raw (pd.DataFrame): Returns a chain of trip and park activities.
        """
        self.trips = trips
        split_overnight_trips = self.user_config["dataparsers"]["split_overnight_trips"]
        self.activities_raw = self._copy_rows(trips=self.trips)
        self.activities_raw = self._add_util_attributes(
            activities_raw=self.activities_raw
        )
        self.activities_raw = self._add_park_act_before_first_trip(
            activities_raw=self.activities_raw, user_config=self.user_config
        )
        self.activities_raw = self._adjust_park_attrs(
            activities_raw=self.activities_raw
        )
        self.activities_raw = self._drop_redundant_columns(
            activities_raw=self.activities_raw
        )
        self.activities_raw = self._remove_next_day_park_acts(
            activities_raw=self.activities_raw
        )
        self.__adjust_park_timestamps()
        self.activities_raw = self._add_next_and_prev_ids(
            activities_raw=self.activities_raw
        )
        self.activities_raw = self._add_first_trip_park_columns(self.activities_raw)
        self.activities_raw = self.__overnight_split_decider(
            split=split_overnight_trips
        )
        self.activities_raw = self._add_timedelta_column(
            activities_raw=self.activities_raw
        )
        self.activities_raw = self._unique_indeces(activities_raw=self.activities_raw)
        logging.info(
            f'Finished activity composition with {self.activities_raw["trip_id"].fillna(0).astype(bool).sum()} trips '
            f'and {self.activities_raw["park_id"].fillna(0).astype(bool).sum()} parking activites.'
        )
        self.activities_raw = self._include_profiles_immobile_vehicles()
        logging.info(
            f'Finished activity composition with {self.activities_raw["trip_id"].fillna(0).astype(bool).sum()} trips '
            f'and {self.activities_raw["park_id"].fillna(0).astype(bool).sum()} parking activites including vehicle which did not perform any trips on the survey date.'
        )
        self.data['activities'] = self.activities_raw
        return self.data

    @staticmethod
    def _copy_rows(trips: pd.DataFrame):
        """
        Adds skeleton duplicate rows for parking activities.

        Args:
            trips (pd.DataFrame): Trip chain data set containing multiple unique_ids

        Returns:
            activities_raw (pd.DataFrame): Raw data set with a copy of each trip added behind the individual trip.
        """
        activities_raw = pd.concat([trips] * 2).sort_index(ignore_index=True)
        activities_raw["park_id"] = activities_raw["trip_id"]
        activities_raw.loc[range(1, len(activities_raw), 2), "trip_id"] = pd.NA
        activities_raw.loc[range(0, len(activities_raw), 2), "park_id"] = pd.NA
        return activities_raw

    @staticmethod
    def _add_util_attributes(activities_raw: pd.DataFrame):
        """
        Adding additional attribute columns with previous and next unique_id.

        Args:
            activities_raw (pd.DataFrame): A raw chain of trip and park activities

        Returns:
            pd.DataFrame: Returns a chain of trip and park activities with additional attributes
        """
        activities_raw["previous_unique_id"] = activities_raw["unique_id"].shift(
            fill_value=0
        )
        activities_raw["is_first_activity"] = (
            activities_raw["previous_unique_id"] != activities_raw["unique_id"]
        )
        activities_raw["next_unique_id"] = activities_raw["unique_id"].shift(
            -1, fill_value=0
        )
        activities_raw["is_last_activity"] = (
            activities_raw["next_unique_id"] != activities_raw["unique_id"]
        )
        return activities_raw

    @staticmethod
    def _add_park_act_before_first_trip(activities_raw: pd.DataFrame, user_config):
        """
        Adds park activities before first trips. Currently, it is assumed that all cars start home.


        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            _type_: _description_
        """
        dataset = user_config["global"]["dataset"]
        new_indeces = activities_raw.index[activities_raw["is_first_activity"]]
        df_add = activities_raw.loc[new_indeces, :]
        df_add["park_id"] = 0
        df_add["purpose_string"] = user_config["dataparsers"][
            "location_park_before_first_trip"
        ][dataset]
        activities_raw.loc[new_indeces, "is_first_activity"] = False
        activities_raw = pd.concat([activities_raw, df_add]).sort_index()
        activities_raw.loc[
            (activities_raw["is_first_activity"]) & (activities_raw["park_id"] == 0),
            "trip_id",
        ] = pd.NA
        return activities_raw

    @staticmethod
    def _adjust_park_attrs(activities_raw: pd.DataFrame) -> pd.DataFrame:
        """
        Sets trip attribute values to zero where trip_id == NaN (i.e. for all parking activities).

        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        activities_raw.loc[
            activities_raw["trip_id"].isna(),
            ["trip_distance", "travel_time", "trip_is_intermodal"],
        ] = pd.NA
        activities_raw["column_from_index"] = activities_raw.index
        activities_raw = activities_raw.sort_values(
            by=["column_from_index", "park_id", "trip_id"]
        )
        return activities_raw

    @staticmethod
    def _drop_redundant_columns(activities_raw: pd.DataFrame) -> pd.DataFrame:
        """
        Removes temporary redundant columns.

        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            activities_raw (pd.DataFrame): _description_
        """
        activities_raw.drop(
            columns=[
                "trip_start_clock",
                "trip_end_clock",
                "trip_start_year",
                "trip_start_month",
                "trip_start_week",
                #"trip_start_hour",
                "trip_start_minute",
                "trip_end_hour",
                "trip_end_minute",
                "previous_unique_id",
                "next_unique_id",
                "column_from_index",
            ],
            inplace=True,
        )
        return activities_raw

    @staticmethod
    def _remove_next_day_park_acts(activities_raw: pd.DataFrame) -> pd.DataFrame:
        """
        Checks for trips across day-limit and removing respective parking activities after ovenight trips.

        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        indeces_multi_day_activities = (
            activities_raw["is_last_activity"]
            & activities_raw["trip_end_next_day"]
            & activities_raw["park_id"]
        )
        unique_ids = activities_raw.loc[indeces_multi_day_activities, "unique_id"]
        trip_ids = (
            activities_raw.loc[
                activities_raw["unique_id"].isin(unique_ids), ["unique_id", "trip_id"]
            ]
            .groupby(by=["unique_id"])
            .max()
        )
        idx = [(i, trip_ids.loc[i].values[0]) for i in trip_ids.index]
        activities_raw = activities_raw.loc[~indeces_multi_day_activities, :]
        acts = activities_raw.copy().set_index(["unique_id", "trip_id"], drop=True)
        acts.loc[idx, "is_last_activity"] = True
        activities_raw = acts.reset_index()
        return activities_raw

    def __adjust_park_timestamps(self):
        """
        Adjust the start and end timestamps of the newly added rows. This is done via range index, that is reset at
        the beginning. First and last activities have to be treated separately since their dates have to match with
        their daily activity chain.
        """
        self.activities_raw = self.activities_raw.reset_index(drop=True)
        park_act_wo_first, park_act_wo_last = self._get_park_acts_wo_first_and_last(activities_raw=self.activities_raw)
        self.activities_raw = self._update_park_start(activities_raw=self.activities_raw,park_act_wo_first=park_act_wo_first)
        self.activities_raw = self._update_park_end(activities_raw=self.activities_raw, park_act_wo_last=park_act_wo_last)
        self.activities_raw = self._update_timestamp_first_park_act(activities_raw=self.activities_raw)
        self.activities_raw = self._update_timestamp_last_park_act(activities_raw=self.activities_raw)
        logging.info("Completed park timestamp adjustments.")

    @staticmethod
    def _get_park_acts_wo_first_and_last(activities_raw: pd.DataFrame) -> pd.DataFrame:
        """
        Returns all parking activities except for the last one (return argument 1) and the first one (return argument
        2)

        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: Parking activity indices without the last one and parking activity indices without the first one
        """
        park_act = ~activities_raw["park_id"].isna()
        park_act = park_act.loc[park_act]
        return park_act.iloc[1:], park_act.iloc[:-1]

    @staticmethod
    def _update_park_start(
        activities_raw: pd.DataFrame, park_act_wo_first: pd.Series
    ) -> pd.DataFrame:
        """
        Updates park start timestamps for newly added rows.

        Args:
            activities_raw (pd.DataFrame): _description_
            park_act_wo_first (pd.Series): _description_

        Returns:
            pd.DataFrame: _description_
        """
        set_timestamp = activities_raw.loc[park_act_wo_first.index - 1, "timestamp_end"]
        set_timestamp.index = activities_raw.loc[
            park_act_wo_first.index, "timestamp_start"
        ].index
        activities_raw.loc[park_act_wo_first.index, "timestamp_start"] = set_timestamp
        return activities_raw

    @staticmethod
    def _update_park_end(
        activities_raw: pd.DataFrame, park_act_wo_last: pd.Series
    ) -> pd.DataFrame:
        """
        Updates park end timestamps for newly added rows.

        Args:
            activities_raw (pd.DataFrame): _description_
            park_act_wo_last (pd.Series): _description_

        Returns:
            pd.DataFrame: _description_
        """
        set_timestamp = activities_raw.loc[
            park_act_wo_last.index + 1, "timestamp_start"
        ]
        set_timestamp.index = activities_raw.loc[
            park_act_wo_last.index, "timestamp_end"
        ].index
        activities_raw.loc[park_act_wo_last.index, "timestamp_end"] = set_timestamp
        return activities_raw

    @staticmethod
    def _update_timestamp_first_park_act(activities_raw: pd.DataFrame) -> pd.DataFrame:
        """
        Updates park end timestamps for last activity in new park rows.

        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        indeces_activities = ~(activities_raw["park_id"].isna()) & (
            activities_raw["is_first_activity"]
        )
        activities_raw.loc[indeces_activities, "timestamp_start"] = replace_vec(
            activities_raw.loc[indeces_activities, "timestamp_end"], hour=0, minute=0
        )
        return activities_raw

    @staticmethod
    def _update_timestamp_last_park_act(activities_raw: pd.DataFrame) -> pd.DataFrame:
        """
        Updates park end timestamps for last activity in new park rows

        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        indeces_activities = ~(activities_raw["park_id"].isna()) & (
            activities_raw["is_last_activity"]
        )
        activities_raw.loc[indeces_activities, "timestamp_end"] = replace_vec(
            activities_raw.loc[indeces_activities, "timestamp_start"], hour=0, minute=0
        ) + pd.Timedelta(1, "d")
        return activities_raw

    @staticmethod
    def _add_next_and_prev_ids(activities_raw: pd.DataFrame) -> pd.DataFrame:
        """
        _summary_

        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        activities_raw.loc[~activities_raw["trip_id"].isna(), "activity_id"] = (
            activities_raw["trip_id"]
        )
        activities_raw.loc[~activities_raw["park_id"].isna(), "activity_id"] = (
            activities_raw["park_id"]
        )
        activities_raw.loc[~activities_raw["is_last_activity"], "next_activity_id"] = (
            activities_raw.loc[:, "activity_id"].shift(-1)
        )
        activities_raw.loc[
            ~activities_raw["is_first_activity"], "previous_activity_id"
        ] = activities_raw.loc[:, "activity_id"].shift(1)
        return activities_raw

    def __overnight_split_decider(self, split: bool):
        """
        Boolean function that differentiates if overnight trips should be split (split==True) or not (split==False).
        In the latter case, overnight trips identified by the variable 'trip_end_next_day' are excluded from the data set.

        Args:
            split (bool): Should trips that end on the consecutive day (not the survey day) be split in two trips in
            such a way that the estimated trip distance the next day is appended in the morning hours of the survey day?
        """
        if split:
            return self.overnight_splitter.split_overnight_trips(
                activities_raw=self.activities_raw
            )
        else:
            self.activities_raw = self._set_overnight_var_false_for_last_act_trip(
                activities_raw=self.activities_raw
            )
            return self._neglect_overnight_trips(activities_raw=self.activities_raw)
        

    @staticmethod
    def _set_overnight_var_false_for_last_act_trip(
        activities_raw: pd.DataFrame,
    ) -> pd.DataFrame:
        """
        Function to treat the edge case of trips being the last activity in the daily activity chain, i.e. trips
        ending exactly at 00:00. They are falsely labelled as overnight trips which is corrected here.

        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        indeces_last_activity_is_trip = (activities_raw["is_last_activity"]) & ~(
            activities_raw["trip_id"].isna()
        )
        idx_last_trip_end_midnight = (
            indeces_last_activity_is_trip
            & (
                activities_raw.loc[
                    indeces_last_activity_is_trip, "timestamp_end"
                ].dt.hour
                == 0
            )
            & (
                activities_raw.loc[
                    indeces_last_activity_is_trip, "timestamp_end"
                ].dt.minute
                == 0
            )
        )
        unique_id_last_trip_end_midnight = activities_raw.loc[
            idx_last_trip_end_midnight, "unique_id"
        ]
        activities_raw.loc[
            activities_raw["unique_id"].isin(unique_id_last_trip_end_midnight.unique()),
            "trip_end_next_day",
        ] = False
        return activities_raw

    @staticmethod
    def _neglect_overnight_trips(activities_raw: pd.DataFrame):
        """
        Removes all overnight trips from the activities data set based on the column 'trip_end_next_day'. Updates
        timestamp end (to 00:00) and is_last_activity for the new last parking activities. Overwrites
        self.activities_raw.

        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            _type_: _description_
        """
        # Column for lastActivity setting later
        activities_raw["next_trip_end_next_day"] = activities_raw[
            "trip_end_next_day"
        ].shift(-1, fill_value=False)

        # Get rid of overnight trips
        indeces_no_overnight_trip = ~(
            activities_raw["trip_end_next_day"].fillna(False)
        )
        activities_raw = activities_raw.loc[indeces_no_overnight_trip, :]

        # Update is_last_activity and timestamp_end variables and clean-up column
        indeces_new_last_activity = activities_raw["next_trip_end_next_day"]
        indeces_new_last_activity = indeces_new_last_activity.fillna(False).astype(bool)
        activities_raw.loc[indeces_new_last_activity, "is_last_activity"] = True
        activities_raw.loc[indeces_new_last_activity, "timestamp_end"] = replace_vec(
            activities_raw.loc[indeces_new_last_activity, "timestamp_start"],
            hour=0,
            minute=0,
        ) + pd.Timedelta(1, "d")
        activities_raw = activities_raw.drop(columns=["next_trip_end_next_day"])
        return activities_raw

    def _add_first_trip_park_columns(self, activities_raw: pd.DataFrame):
        activities_raw = self._is_first_trip(activities_raw=activities_raw)
        activities_raw = self._is_first_park_activity(activities_raw=activities_raw)
        return activities_raw

    @staticmethod
    def _is_first_trip(activities_raw: pd.DataFrame):
        acts_idx = activities_raw.set_index(['unique_id', 'trip_id'])
        first_trip = activities_raw[['unique_id', 'trip_id']].groupby(by='unique_id').min(numeric_only=True)
        first_trip['is_first_trip'] = True
        first_trip = first_trip.set_index('trip_id', append=True)
        acts_idx['is_first_trip'] = first_trip  # index comprehension
        acts_idx['is_first_trip'] = acts_idx['is_first_trip'].fillna(value=False)
        return acts_idx.reset_index()

    @staticmethod
    def _is_first_park_activity(activities_raw: pd.DataFrame):
        acts_idx = activities_raw.set_index(['unique_id', 'park_id'])
        first_park = activities_raw[['unique_id', 'park_id']].groupby(by='unique_id').min(numeric_only=True)
        first_park['is_first_park_activity'] = True
        first_park = first_park.set_index('park_id', append=True)
        acts_idx['is_first_park_activity'] = first_park  # index comprehension
        acts_idx['is_first_park_activity'] = acts_idx['is_first_park_activity'].fillna(value=False)
        return acts_idx.reset_index()

    @staticmethod
    def _add_timedelta_column(activities_raw: pd.DataFrame):
        """
        Adds column keeping the length information of the activity as a pandas time_delta.

        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            _type_: _description_
        """
        activities_raw["time_delta"] = (
            activities_raw["timestamp_end"] - activities_raw["timestamp_start"]
        )
        return activities_raw

    @staticmethod
    def _unique_indeces(activities_raw: pd.DataFrame):
        """
        _summary_

        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            _type_: _description_
        """
        activities_raw.reset_index(inplace=True, drop=True)  # Due to copying and appending rows, the index has to be reset
        return activities_raw

    def _calculate_vehicle_numbers(self, data):
        number_mobile_vehicles = [
            len(_)
            for _ in data
            .unique_id.unique()
            .to_list()
        ]
        number_mobile_vehicles = {i: number_mobile_vehicles[i - 1] for i in range(1, 8)}
        if self.user_config["dataparsers"]["subset_vehicle_segment"]:
            vehicle_segment = self.user_config["dataparsers"]["vehicle_segment"][self.dataset]
            if self.dataset == "KiD":
                percentages_immobile_vehicles = self.user_config["postprocessor"]["immobile_vehicles"][self.dataset][vehicle_segment]
            else:
                percentages_immobile_vehicles = self.user_config["postprocessor"]["immobile_vehicles"][self.dataset]
        else:
            #TODO: line below needed?
            percentages_immobile_vehicles = self.user_config["postprocessor"]["immobile_vehicles"][self.dataset]
        number_immobile_vehicles = {
            key: number_mobile_vehicles[key]
            / (100 - percentages_immobile_vehicles[key])
            * percentages_immobile_vehicles[key]
            for key in number_mobile_vehicles.keys()
            & percentages_immobile_vehicles.keys()
        }
        number_immobile_vehicles = {
            key: round(number_immobile_vehicles[key])
            for key in number_immobile_vehicles.keys()
        }
        total_amount_vehicles = {
            key: number_mobile_vehicles[key] + number_immobile_vehicles[key]
            for key in number_mobile_vehicles.keys() & number_immobile_vehicles.keys()
        }
        self.vehicle_numbers = {
            "mobile_vehicles": number_mobile_vehicles,
            "immobile_vehicles": number_immobile_vehicles,
            "total_amount_vehicle": total_amount_vehicles,
        }
        return self.vehicle_numbers
 
    def _calculate_seasonal_vehicle_numbers(self, data):
        for season in self.seasons:
            self.seasonal_vehicle_numbers[season] = self._calculate_vehicle_numbers(data)
        return self.seasonal_vehicle_numbers

    def _create_structure_with_weekday(self, vehicle_numbers):
        immobile_vehicles = pd.DataFrame()
        for key, value in vehicle_numbers.items():
            immobile_vehicles_df_structure = pd.DataFrame(np.nan, index=range(value), columns=self.activities_raw.columns)
            immobile_vehicles_df_structure["trip_start_weekday"] = int(key)
            immobile_vehicles = immobile_vehicles.append(immobile_vehicles_df_structure).reset_index(drop=True)
        immobile_vehicles.loc[:, "weekday_string"] = immobile_vehicles.loc[:, "trip_start_weekday"].replace(
            self.dev_config["dataparsers"]["replacements"][self.dataset]["trip_start_weekday"]
        )
        return immobile_vehicles

    def _create_unique_id(self):
        unique_id_to_exclude = self.activities_raw.unique_id.unique()
        # Generate a pool of random IDs (larger than the DataFrame size)
        pool_size = len(self.immobile_vehicles) + len(unique_id_to_exclude)
        #TODO: check if ids are different between datasets
        random_ids = np.random.choice(range(1000000, 10000000), size=pool_size, replace=False)
        # Filter out IDs that are in the exclusion list
        filtered_ids = [id for id in random_ids if id not in unique_id_to_exclude]
        # Assign the filtered IDs to the DataFrame
        self.immobile_vehicles['unique_id'] = filtered_ids[:len(self.immobile_vehicles)]
        return self.immobile_vehicles

    def _set_vehicle_segment(self):
        if self.user_config["dataparsers"]["subset_vehicle_segment"]:
            if self.dataset == "KiD" or self.dataset == "VF":
                self.immobile_vehicles['vehicle_segment_string'] = self.user_config["dataparsers"]["vehicle_segment"][self.dataset]
        return self.immobile_vehicles

    def _set_columns_to_one(self):
        self.immobile_vehicles['park_id'] = 1
        self.immobile_vehicles['activity_id'] = 1
        #self.immobile_vehicles['is_first_parking_act'] = 1
        return self.immobile_vehicles

    def _set_boolean_columns(self):
        self.immobile_vehicles['is_first_activity'] = True
        self.immobile_vehicles['is_last_activity'] = True
        self.immobile_vehicles['is_first_park_activity'] = True
        self.immobile_vehicles['is_first_trip'] = False
        return self.immobile_vehicles

    def _create_timestamps_and_timedelta(self):
        today = datetime.date.today()
        start_timestamp = pd.Timestamp(today)
        end_timestamp = pd.Timestamp(today)
        self.immobile_vehicles['timestamp_start'] = start_timestamp
        self.immobile_vehicles['timestamp_end'] = end_timestamp
        #TODO: check in diarybuilder if activities where start and end times are equal get deleted, change with check on timedelta maybe?
        duration_24h = datetime.timedelta(hours=24)
        self.immobile_vehicles["time_delta"] = duration_24h
        return self.immobile_vehicles

    def _include_trip_purpose(self):
        #TODO: assign trip purpose based on probability distribution?
        #TODO: move this to config and add edge case for other datasets
        self.immobile_vehicles['purpose_string'] = "HOME"
        return self.immobile_vehicles

    def _include_area_type(self):
        self.immobile_vehicles['area_type'] = "area_type"
        #TODO: decide how to treat area type
        return self.immobile_vehicles

    def _add_parking_activity_immobile_vehicles(self):
        if self.season == "all" and self.user_config["global"]["consider_temperature_cycle_dependency"]["annual"] is True:
            for season in self.seasons:
                immobile_season = self._create_structure_with_weekday(self.seasonal_vehicle_numbers[season]['immobile_vehicles'])
                immobile_season['season'] = season
                self.immobile_vehicles = pd.concat([self.immobile_vehicles, immobile_season])
        elif pd.Series(self.season).isin(self.seasons)[0]:
            self.immobile_vehicles = self._create_structure_with_weekday(self.vehicle_numbers['immobile_vehicles'])
            self.immobile_vehicles['season'] = self.season
        else:
            self.immobile_vehicles = self._create_structure_with_weekday(self.vehicle_numbers['immobile_vehicles'])
            self.immobile_vehicles['season'] = 'not assigned'
        self.immobile_vehicles = self._create_unique_id()
        self.immobile_vehicles = self._set_vehicle_segment()
        self.immobile_vehicles = self._set_columns_to_one()
        self.immobile_vehicles = self._set_boolean_columns()
        self.immobile_vehicles = self._create_timestamps_and_timedelta()
        self.immobile_vehicles = self._include_trip_purpose()
        self.immobile_vehicles = self._include_area_type()
        self.activities_raw = pd.concat([self.activities_raw, self.immobile_vehicles]).reset_index(drop=True)
        return self.activities_raw

    def _include_profiles_immobile_vehicles(self):
        if self.season == "all" and self.user_config["global"]["consider_temperature_cycle_dependency"]["annual"] is True:
            self.data['vehicles_number'] = self._calculate_seasonal_vehicle_numbers(self.activities_raw.groupby(by=["trip_start_weekday", "season"]))
        else:
            self.data['vehicles_number'] = self._calculate_vehicle_numbers(self.activities_raw.groupby(by="trip_start_weekday"))
        self.activities_raw = self._add_parking_activity_immobile_vehicles()
        return self.activities_raw



class OvernightSplitter:
    def __init__(self):
        """
        _summary_
        """
        self.activities_raw = None  # internally used until merge of morning splits
        self.activities = None  # internally used for merged data

    def split_overnight_trips(self, activities_raw: pd.DataFrame) -> pd.DataFrame:
        """
        Wrapper function for treating edge case trips ending not in the 24 hours of the survey day but stretch
        to the next day. Those overnight are split up into an evening trip at the regular survey day and a
        morning trip at the next day. Trip distances are split according to the time the person spent on that trip.
        E.g. if a trip lasts from 23:00 to 2:00 the next day and 100 km, the split-up evening trip will last from
        23:00 to 00:00 of the survey day and 33 km and the morning trip from 00:00 to 2:00 and 66 km. In a next step,
        the morning trip is appended to the survey day in the first hours.

        Here, different edge cases occur.
        Edge case 1 (N=5 in MiD17): For trips that overlap with night (early morning) trips at the survey day, e.g. from
        0:30 to 1:00 for the above mentioned example, the morning part of the split overnight trip is completely
        disregarded.
        Edge case 2 (N=3 in MiD17): When overnight mornging split-trips end exactly at the time where the first trip of
        the survey day starts (2:00 in the example), both trips are consolidated to one trip with all attributes of the
        survey trip.
        These edge cases are documented and quantified in issue #358 'Sum of all distances of dataParser at end equals
        sum of all distances after filtering'.

        Args:
            activities_raw (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        self.activities_raw = activities_raw

        # Split overnight trips and adjust last trip variables accordingly
        is_overnight_trip, overnight_trips_add, ON_uids = (
            self.__get_overnight_activities()
        )
        ON_trips_add_timestamp = self.__adjust_overnight_timestamps(
            trips=overnight_trips_add
        )
        self.__set_last_activities_end_timestamp_to_zero()

        # Adjust morning split trips (trip_id=0) and their variables
        morning_trips = self.__set_overnight_trip_id_to_zero(
            trips=ON_trips_add_timestamp
        )
        morning_trips = self.__adjust_morning_trip_distance(
            overnightTrips=overnight_trips_add, morning_trips=morning_trips
        )

        self.__adjust_evening_trip_distance(
            morning_trips=morning_trips, is_overnight_trip=is_overnight_trip
        )
        morning_trips = self.__set_morning_trips_first_last_acts(
            morning_trips=morning_trips
        )
        is_prev_first_acts = self.__get_prev_first_act(
            morning_trips=morning_trips
        )  # Parking activities that are still first activities in the activities data set
        morning_trips_no_overlap, is_prev_first_acts = (
            self.__neglect_overlap_morning_trips(
                morning_trips=morning_trips, is_prev_first_acts=is_prev_first_acts
            )
        )  # neglect morning split trips that overlap with first trip
        self.__set_first_parking_timestamp_start(
            morning_trips=morning_trips_no_overlap,
            is_overnight_trip=is_overnight_trip,
            is_prev_first_acts=is_prev_first_acts,
        )
        morning_trips_to_add = self.__set_morning_split_act_id_zero(
            morning_trips=morning_trips_no_overlap
        )
        self.__set_is_first_trip_false(ids=ON_uids)
        self.__add_morning_trips(morning_trips=morning_trips_to_add)
        # self.__remove_first_parking_act() DEPRECATED
        neglected_trips = self.__merge_adjacent_trips()
        # Implement DELTA mileage check of overnight morning split trip distances
        self.__check_and_assert(neglected_trips=neglected_trips)
        self.__clean_up_columns()
        self.__sort_activities()

        return self.activities

    def __get_overnight_activities(self) -> tuple[pd.Series, pd.DataFrame, pd.Series]:
        """
        _summary_

        Returns:
            tuple[pd.Series, pd.DataFrame]: _description_
        """
        indeces_overnight_actvities = (
            self.activities_raw["is_last_activity"]
            & self.activities_raw["trip_end_next_day"]
            & ~(
                (self.activities_raw["timestamp_end"].dt.hour == 0)
                & (
                    self.activities_raw["timestamp_end"].dt.minute == 0
                )  # assure that the overnight trip does
            )
        )  # not exactly end at 00:00
        overnight_activities = self.activities_raw.loc[indeces_overnight_actvities, :]
        return (
            indeces_overnight_actvities,
            overnight_activities,
            overnight_activities["unique_id"],
        )

    def __adjust_overnight_timestamps(self, trips: pd.DataFrame) -> pd.DataFrame:
        """
        _summary_

        Args:
            trips (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        tripsRes = trips.copy()
        tripsRes["timestamp_end"] = tripsRes.loc[:, "timestamp_end"] - pd.Timedelta(
            1, "d"
        )
        tripsRes["timestamp_start"] = replace_vec(
            tripsRes.loc[:, "timestamp_end"], hour=0, minute=0
        )
        return tripsRes

    def __set_last_activities_end_timestamp_to_zero(self):
        """
        _summary_
        """
        # Set timestamp end of evening part of overnight trip split to 00:00
        self.activities_raw.loc[
            self.activities_raw["is_last_activity"], "timestamp_end"
        ] = replace_vec(
            self.activities_raw.loc[
                self.activities_raw["is_last_activity"], "timestamp_end"
            ],
            hour=0,
            minute=0,
        )

    def __set_overnight_trip_id_to_zero(self, trips: pd.DataFrame) -> pd.DataFrame:
        """
        _summary_

        Args:
            trips (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        trips["trip_id"] = 0
        trips["activity_id"] = 0
        trips["previous_activity_id"] = pd.NA

        # Update next activity ID

        # SUPPOSEDLY DEPRECATED / OVERLY COMPLEX PART
        # unique_id = trips["unique_id"]
        # act_idx = self.activities_raw["unique_id"].isin(unique_id) & self.activities_raw["is_first_activity"]
        # trips["next_activity_id"] = self.activities_raw.loc[act_idx, "activity_id"]

        # overnight morning splits are always first activities and thus the ones before park activities with park_id=0
        trips["next_activity_id"] = 0

        # Update previous activity ID of previously first activity
        act_idx = (
            self.activities_raw["unique_id"].isin(trips["unique_id"])
            & self.activities_raw["is_first_activity"]
        )
        self.activities_raw.loc[act_idx, "previous_activity_id"] = 0
        return trips

    def __adjust_morning_trip_distance(
        self, overnightTrips: pd.DataFrame, morning_trips: pd.DataFrame
    ) -> pd.DataFrame:
        """
        _summary_

        Args:
            overnightTrips (pd.DataFrame): _description_
            morning_trips (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        # Splitting the total distance to morning and evening trip time-share dependent
        morning_trips["timedelta_total"] = (
            overnightTrips["timestamp_end"] - overnightTrips["timestamp_start"]
        )
        morning_trips["timedelta_morning"] = (
            morning_trips["timestamp_end"] - morning_trips["timestamp_start"]
        )
        morning_trips["time_share_morning"] = (
            morning_trips["timedelta_morning"] / morning_trips["timedelta_total"]
        )
        morning_trips["time_share_evening"] = (
            morning_trips["timedelta_total"] - morning_trips["timedelta_morning"]
        ) / morning_trips["timedelta_total"]
        morning_trips["total_trip_distance"] = morning_trips["trip_distance"]
        morning_trips["trip_distance"] = (
            morning_trips["time_share_morning"] * morning_trips["total_trip_distance"]
        )
        return morning_trips

    def __adjust_evening_trip_distance(
        self, morning_trips: pd.DataFrame, is_overnight_trip: pd.Series
    ):
        """
        _summary_

        Args:
            morning_trips (pd.DataFrame): _description_
            is_overnight_trip (pd.Series): _description_
        """
        self.activities_raw.loc[is_overnight_trip, "trip_distance"] = (
            morning_trips["time_share_evening"] * morning_trips["total_trip_distance"]
        )

    def __set_morning_trips_first_last_acts(
        self, morning_trips: pd.DataFrame
    ) -> pd.DataFrame:
        """
        _summary_

        Args:
            morning_trips (pd.DataFrame): _description_
        """
        # Setting first and last activities
        morning_trips["is_first_activity"] = True
        morning_trips["is_first_trip"] = True
        morning_trips["is_last_activity"] = False
        return morning_trips

    def __get_prev_first_act(self, morning_trips: pd.DataFrame):
        """
        _summary_

        Args:
            morning_trips (pd.DataFrame): _description_

        Returns:
            _type_: _description_
        """
        return (
            self.activities_raw["unique_id"].isin(morning_trips["unique_id"])
            & self.activities_raw["is_first_activity"]
        )

    def __neglect_overlap_morning_trips(
        self, morning_trips: pd.DataFrame, is_prev_first_acts: pd.DataFrame
    ) -> tuple[pd.DataFrame, pd.Series]:
        """
        _summary_

        Args:
            morning_trips (pd.DataFrame): _description_
            is_prev_first_acts (pd.DataFrame): _description_

        Returns:
            tuple[pd.DataFrame, pd.Series]: _description_
        """
        # Option 1 of treating overlaps: After concatenation in the end

        on_uids = morning_trips["unique_id"]  # overnight day unique ids
        first_park_end = self.activities_raw.loc[
            (self.activities_raw["unique_id"].isin(on_uids))
            & (self.activities_raw["is_first_park_activity"]),
            "timestamp_end",
        ].copy()
        first_park_end.index = morning_trips.index  # Adjust index for comparison

        # Filter out morning parts of overnight trip split for persons that already have morning trips in that period
        neglect_overnight = first_park_end < morning_trips["timestamp_end"]
        morning_trips_no_overlap = morning_trips.loc[~neglect_overnight, :]

        # Filter out neglected activities from prev_first_acts accordingly
        indeces_neglect_overnight = neglect_overnight
        indeces_neglect_overnight.index = is_prev_first_acts[
            is_prev_first_acts
        ].index  # Align index for filtering
        indeces_neglect_overnight = indeces_neglect_overnight[indeces_neglect_overnight]
        is_prev_first_acts[indeces_neglect_overnight.index] = False

        return morning_trips_no_overlap, is_prev_first_acts

    def __set_first_parking_timestamp_start(
        self,
        morning_trips: pd.DataFrame,
        is_overnight_trip: pd.Series,
        is_prev_first_acts: pd.Series,
    ) -> pd.DataFrame:
        """
        Sets start timestamp of previously first activity (parking) to end timestamp of morning split of overnight trip.

        Args:
            morning_trips (pd.DataFrame): _description_
            is_overnight_trip (pd.Series): _description_
            is_prev_first_acts (pd.DataFrame): _description_

        Returns:
            pd.DataFrame: _description_
        """
        timestamp_new = morning_trips["timestamp_end"].copy()
        timestamp_new.index = self.activities_raw.loc[
            is_prev_first_acts, "timestamp_start"
        ].index
        self.activities_raw.loc[is_prev_first_acts, "timestamp_start"] = timestamp_new
        self.activities_raw.loc[is_prev_first_acts, "is_first_activity"] = False

    def __set_morning_split_act_id_zero(
        self, morning_trips: pd.DataFrame
    ) -> pd.DataFrame:
        """
        The first parking activity id is always 0. The exception of morning split trip and
        first trip exactly touching is treated later.

        Args:
            morning_trips (pd.DataFrame): Morning split trips as rows with venco.py variables
                in columns.

        Returns:
            pd.DataFrame: The morning split trips with next_activity_id set to 0
        """
        m = morning_trips.copy()
        m["next_activity_id"] = 0
        return m

    def __set_is_first_trip_false(self, ids: pd.Series):
        self.activities_raw.loc[
            (self.activities_raw["unique_id"].isin(ids))
            & (self.activities_raw["is_first_trip"]),
            "is_first_trip",
        ] = False

    def __add_morning_trips(self, morning_trips: pd.DataFrame):
        """
        Appends overnight morning trips.

        Args:
            morning_trips (pd.DataFrame): _description_
        """
        self.activities = pd.concat(
            [self.activities_raw, morning_trips], ignore_index=True
        )

    # DEPRECATED
    def __remove_first_parking_act(self):
        """
        Removes first parking activities for persons where first activity is a trip (starting at 00:00).
        """
        first_park_acts = self.activities.loc[
            self.activities["is_first_park_activity"], :
        ]
        first_trip_acts = self.activities.loc[self.activities["is_first_trip"], :]
        first_trip_acts.index = first_park_acts.index  # Aligning trip indices
        indeces_park_timestamp = (
            first_park_acts["timestamp_start"] == first_trip_acts["timestamp_start"]
        )
        self.activities = self.activities.drop(
            indeces_park_timestamp[indeces_park_timestamp].index
        )

        # After removing first parking, set first trip to first activity
        self.activities.loc[
            (
                self.activities["unique_id"].isin(
                    first_park_acts.loc[indeces_park_timestamp, "unique_id"]
                )
            )
            & (self.activities["trip_id"] == 1),
            "is_first_activity",
        ] = True

    def __merge_adjacent_trips(self):
        """
        Consolidate overnight morning trips and first trips for the edge case where morning trips of next day
        end exactly at the beginning of the first trip of the survey day. In this case, the morning split of the
        overnight trip is neglected and the beginning of the first trip is set to 00:00. In the MiD17 data set, there
        were 3 occurences of this case all with end times of the overnight trip between 00:00 and 01:00.
        """

        # identify uids where timestamp_end (trip_id == 0) and timestamp start (trip_id == 1) are equal
        # and from those uids the trips with trip_id == 0. Those should be neglected
        self.activities, neglected_trips, remaining_first_trips = (
            self.__neglect_morning_splits()
        )

        # set timestamp_start to 00:00 of previously first trip and previous activity id to pd.NA
        self.__update_consolidated_act(
            neglected_trips=neglected_trips, remaining_trips=remaining_first_trips
        )

        return neglected_trips

    def __check_and_assert(self, neglected_trips: pd.DataFrame):
        """
        _summary_
        """
        # Calculates the neglected trip distances from overnight split trips with regular morning trips
        total_distance = self.activities.loc[
            ~self.activities["trip_id"].isna(), "trip_distance"
        ].sum()
        neglected_trip_distance = neglected_trips["trip_distance"].sum()
        ratio = neglected_trip_distance / total_distance
        logging.info(
            f"From {round(total_distance, 2)} km total mileage in the dataset after filtering, "
            f"{round((ratio * 100), 2)} % were cropped because they corresponded to split-trips from overnight trips."
        )
        assert ratio < 0.01

    def __neglect_morning_splits(
        self,
    ) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Identifies the household person IDs that should be neglected.
        """
        unique_ids_overnight = self.activities.loc[
            self.activities["trip_id"] == 0, "unique_id"
        ]
        activities = self.activities.loc[
            self.activities["unique_id"].isin(unique_ids_overnight), :
        ]

        first_trips = activities.loc[activities["trip_id"] == 0, :]
        # Next trip after morning part of overnight split
        next_trips = activities.loc[
            (activities["previous_activity_id"] == 0) & (activities["park_id"].isna()),
            :,
        ]
        first_park_activities = activities.loc[activities["park_id"] == 0, :]

        # Timestamp comparison using morning trip index because those should eventually be neglected
        bool = (
            first_trips["timestamp_end"].values == next_trips["timestamp_start"].values
        )
        neglect_trips_idx = first_trips.loc[bool, :].index
        neglect_park_idx = first_park_activities.loc[bool, :].index
        neglect_idx = neglect_trips_idx.union(neglect_park_idx)

        # supposedly DEPRECATED
        # remain_trips_idx = next_trips.loc[bool, :].index

        neglected_trips = first_trips[bool]
        remaining_first_trips = next_trips[bool]

        if any(bool):
            remaining_activities = self.activities.loc[
                ~self.activities.index.isin(neglect_idx), :
            ]
        else:
            remaining_activities = self.activities

        return remaining_activities, neglected_trips, remaining_first_trips

    # DEPRECATED
    def __neglect_zero_trip_id_from_activities(self, id_neglect: pd.Series):
        """
        Filters out the activities with the given hhpid and trip_id 0.

        Args:
            id_neglect (pd.Series): _description_
        """
        neglect = (self.activities["unique_id"].isin(id_neglect)) & (
            self.activities["trip_id"] == 0
        )
        self.activities = self.activities.loc[~neglect, :]

    def __update_consolidated_act(
        self, neglected_trips: pd.DataFrame, remaining_trips: pd.DataFrame
    ):
        """
        Sets the start timestamp of the firstActivity of all hhpids given as argument to 00:00. Additionally
        the previous_activity_id is set to pd.NA._summary_

        Args:
            id_neglect (pd.Series): _description_
        """
        idx = remaining_trips.index

        # Adjust timestamp_start
        self.activities.loc[idx, "timestamp_start"] = replace_vec(
            self.activities.loc[idx, "timestamp_start"],
            hour=0,
            minute=0,
        )

        # Add neglected trip distance
        neglected_trips.index = idx
        self.activities.loc[idx, "trip_distance"] = (
            self.activities.loc[idx, "trip_distance"] + neglected_trips["trip_distance"]
        )

        # Set purpose to morning split trip purpose
        self.activities.loc[idx, "trip_purpose"] = neglected_trips["trip_purpose"]

        # Set activity id booleans
        self.activities.loc[idx, "is_first_activity"] = True
        self.activities.loc[idx, "is_first_trip"] = True
        self.activities.loc[idx, "previous_activity_id"] = pd.NA

        # ToDo: Is the park activity in between adjacent trips already deleted?

    def __clean_up_columns(self):
        keep_col_bool = ~self.activities.columns.isin(
            [
                "timedelta_total",
                "timedelta_morning",
                "time_share_morning",
                "time_share_evening",
                "total_trip_distance",
            ]
        )
        self.activities = self.activities.loc[:, keep_col_bool]

    def __sort_activities(self):
        """
        Sorts activities according to unique_id and timestamp_start column values.
        """
        self.activities = self.activities.sort_values(
            by=["unique_id", "timestamp_start"]
        )
    
